import asyncio
import logging
import socket
from typing import Any
import faust
from mode import Service
from .app import get_error_topic
from .models import Error


class LogPusher(Service):

    app: faust.App
    queue: asyncio.Queue

    def __init__(self, app: faust.App, **kwargs: Any) -> None:
        self.app = app
        self.queue = asyncio.Queue()
        super().__init__(**kwargs)

    def put(self, error: Error) -> None:
        self.queue.put_nowait(error)

    @Service.task
    async def _flush(self) -> None:
        app = self.app
        queue = self.queue
        error_topic = get_error_topic(app)
        while not self.should_stop:
            error = await queue.get()
            error.app_id = app.conf.id
            await error_topic.send(value=error)


class LogHandler(logging.Handler):

    app: faust.App

    def __init__(self, app: faust.App, *args: Any, **kwargs: Any) -> None:
        self.app = app
        super().__init__(*args, **kwargs)

    def emit(self, record: logging.LogRecord) -> None:
        self.format(record)
        return self._emit(record)

    def _emit(self, record: logging.LogRecord) -> None:
        no_alert = getattr(record, 'no_alert', False)
        if not no_alert and record.levelno >= logging.ERROR:
            stack = getattr(record, 'stack', None)
            traceback = None
            if stack is True:
                traceback = '\n'.join(traceback.format_stack())
            error = Error(
                message=record.getMessage(),
                format=str(record.msg),
                traceback=traceback,
                logger=record.name,
                hostname=socket.gethostname(),
                filename=record.filename,
                module=record.module,
                lineno=record.lineno,
                severity=record.levelname,
                timestamp=record.created,
            )
            self.app.logpusher.put(error)



from faust.utils import venusian
from mode.utils.mocks import Mock, patch


def test_attach():
    callable = Mock(name='callable')
    category = 'category'
    with patch('faust.utils.venusian._attach') as _attach:
        with patch('faust.utils.venusian._on_found') as _on_found:
            venusian.attach(callable, category)
            _attach.assert_called_with(callable, _on_found, category=category)


def test_on_found():
    venusian._on_found(Mock(name='scanner'), 'name', Mock(name='obj'))



import pytest
from faust.exceptions import SecurityError
from faust.models.tags import Secret, Sensitive


class test_Sensitive:

    @pytest.fixture
    def typ(self):
        return Sensitive[str]

    @pytest.fixture
    def v(self, *, typ):
        return typ('value')

    def test_repr(self, *, v):
        assert repr(v)
        assert 'value' not in repr(v)

    def test_str(self, *, v):
        with pytest.raises(SecurityError):
            str(v)

    def test_nested(self, *, v, typ):
        with pytest.raises(SecurityError):
            typ(v)


class test_Secret(test_Sensitive):

    @pytest.fixture
    def typ(self):
        return Secret[str]

    def test_str(self, *, v):
        assert str(v) == v.mask



from datetime import timedelta, timezone
from faust.livecheck.models import State


class test_State:

    def test_is_ok(self):
        assert State.INIT.is_ok()
        assert State.PASS.is_ok()
        assert State.SKIP.is_ok()
        assert not State.FAIL.is_ok()
        assert not State.ERROR.is_ok()
        assert not State.TIMEOUT.is_ok()
        assert not State.STALL.is_ok()


class test_TestExecution:

    def test_ident(self, *, execution):
        assert execution.ident

    def test_short_ident(self, *, execution):
        assert execution.shortident

    def test_now(self, *, execution):
        assert execution._now()
        assert execution._now().tzinfo is timezone.utc

    def test_human_date__issued_today(self, *, execution):
        execution.was_issued_today = True
        assert execution.human_date

    def test_human_date__issued_before_today(self, *, execution):
        execution.was_issued_today = False
        assert execution.human_date

    def test_was_issued_today(self, *, execution):
        execution.timestamp = execution._now()
        assert execution.was_issued_today

    def test_was_issued_today__not(self, *, execution):
        execution.timestamp = execution._now() - timedelta(days=3)
        assert not execution.was_issued_today

    def test_is_expired(self, *, execution):
        execution.expires = execution._now() - timedelta(hours=1)
        assert execution.is_expired

    def test_is_expired__not(self, *, execution):
        execution.expires = execution._now() + timedelta(hours=1)
        assert not execution.is_expired



import asyncio
from collections import defaultdict
import faust
import pytest
from faust import joins
from faust.exceptions import Skip
from mode.utils.contexts import ExitStack
from mode.utils.mocks import AsyncMock, Mock, patch
from sortedcontainers import SortedSet
from t.helpers import new_event


class Model(faust.Record):
    foo: str


class test_Stream:

    @pytest.fixture()
    def stream(self, *, app):
        return app.stream(app.channel())

    def test_join(self, *, stream):
        s2 = stream.join(Model.foo)
        assert s2.join_strategy
        assert isinstance(s2.join_strategy, joins.RightJoin)
        assert s2.join_strategy.stream is stream
        assert s2.join_strategy.fields == {Model.foo.model: Model.foo}

    def test_left_join(self, *, stream):
        s2 = stream.left_join(Model.foo)
        assert s2.join_strategy
        assert isinstance(s2.join_strategy, joins.LeftJoin)
        assert s2.join_strategy.stream is stream
        assert s2.join_strategy.fields == {Model.foo.model: Model.foo}

    def test_inner_join(self, *, stream):
        s2 = stream.inner_join(Model.foo)
        assert s2.join_strategy
        assert isinstance(s2.join_strategy, joins.InnerJoin)
        assert s2.join_strategy.stream is stream
        assert s2.join_strategy.fields == {Model.foo.model: Model.foo}

    def test_outer_join(self, *, stream):
        s2 = stream.outer_join(Model.foo)
        assert s2.join_strategy
        assert isinstance(s2.join_strategy, joins.OuterJoin)
        assert s2.join_strategy.stream is stream
        assert s2.join_strategy.fields == {Model.foo.model: Model.foo}

    @pytest.mark.asyncio
    async def test_on_merge__with_join_strategy(self, *, stream):
        join = stream.join_strategy = Mock(process=AsyncMock())
        assert (await stream.on_merge('v')) is join.process.coro.return_value

    def test_combine__finalized(self, *, stream):
        stream._finalized = True
        assert stream.combine(Mock()) is stream

    def test_group_by__finalized(self, *, stream):
        stream._finalized = True
        assert stream.group_by(Mock()) is stream

    def test_through__finalized(self, *, stream):
        stream._finalized = True
        assert stream.through(Mock()) is stream

    def test__set_current_event(self, *, stream):
        with patch('faust.streams._current_event') as ce:
            stream._set_current_event(None)
            assert stream.current_event is None
            ce.set.assert_called_once_with(None)
            ce.set.reset_mock()

            with patch('weakref.ref') as ref:
                event = Mock(name='event')
                stream._set_current_event(event)
                assert stream.current_event is event
                ce.set.assert_called_once_with(ref.return_value)
                ref.assert_called_once_with(event)

    @pytest.mark.asyncio
    async def test__format_key(self, *, stream):
        value = Model('val')
        assert (await stream._format_key(Model.foo, value)) == 'val'

    @pytest.mark.asyncio
    async def test__format_key__callable(self, *, stream):
        async def format_key(value):
            return value.foo
        value = Model('val')
        assert (await stream._format_key(format_key, value)) == 'val'

    @pytest.mark.asyncio
    async def test_groupby__field_descriptor(self, *, stream, app):
        async with stream:
            channel = app.channel()
            stream.prefix = 'agent-foo'
            s2 = stream.group_by(Model.foo, topic=channel)

            assert not s2._processors
            assert len(stream._processors) == 1

            repartition = stream._processors[0]

            stream.current_event = None
            with pytest.raises(RuntimeError):
                await repartition(Model('foo'))

            stream.current_event = Mock(name='event', forward=AsyncMock())
            await repartition(Model('foo'))
            stream.current_event.forward.assert_called_once_with(
                channel, key='foo')

    @pytest.mark.asyncio
    async def test_echo(self, *, stream, app):
        channel = Mock(name='channel', send=AsyncMock())
        s2 = stream.echo(channel)
        assert len(s2._processors) == 1
        echoing = s2._processors[0]

        await echoing('val')
        channel.send.assert_called_once_with(value='val')

    @pytest.mark.asyncio
    @pytest.mark.allow_lingering_tasks(count=1)
    async def test_aiter_tracked(self, *, stream, app):
        print('TEST AITER TRACKED')
        event = await self._assert_stream_aiter(
            stream,
            side_effect=None,
            raises=None,
        )
        if not event.ack.called:
            assert event.message.refcount == 0
            assert event.message.acked
        else:
            event.ack.assert_called_once_with()

    @pytest.mark.asyncio
    @pytest.mark.allow_lingering_tasks(count=1)
    async def test_aiter_tracked__CancelledError(self, *, stream, app):
        event = await self._assert_stream_aiter(
            stream,
            side_effect=asyncio.CancelledError(),
            raises=asyncio.CancelledError,
        )
        if not event.ack.called:
            assert event.message.refcount == 0
            assert event.message.acked
        else:
            event.ack.assert_called_once_with()

    async def _assert_stream_aiter(self, stream,
                                   side_effect=None,
                                   raises=None):
        sentinel = object()
        app = stream.app
        app.consumer = Mock(name='app.consumer')
        app.consumer._committed_offset = defaultdict(lambda: -1)
        app.consumer._acked_index = defaultdict(set)
        app.consumer._acked = defaultdict(SortedSet)
        app.consumer._n_acked = 0
        app.flow_control.resume()
        app.topics._acking_topics.add('foo')
        event = new_event(app, topic='foo', value=32)
        event.message.tracked = False
        event.ack = Mock(name='event.ack')
        channel = app.channel()

        got_sentinel = asyncio.Event()

        @app.agent(channel)
        async def agent(stream):
            with ExitStack() as stack:
                if raises:
                    stack.enter_context(pytest.raises(raises))
                async for value in stream:
                    if value is sentinel:
                        got_sentinel.set()
                        break
                    assert value is event.value
                    if side_effect:
                        got_sentinel.set()
                        raise side_effect

        s: faust.StreamT = None
        async with agent as _agent:
            s = list(_agent._actors)[0].stream.get_active_stream()
            await s.channel.put(event)
            await s.channel.put(new_event(app, topic='bar', value=sentinel))
            s._on_message_in = Mock()
            await got_sentinel.wait()
            await asyncio.sleep(0)
            await asyncio.sleep(0)
            await asyncio.sleep(0)
        s._on_message_in.assert_called_once_with(
            event.message.tp,
            event.message.offset,
            event.message,
        )
        return event

    @pytest.mark.asyncio
    async def test_ack(self, *, stream):
        event = Mock()
        event.ack.return_value = True
        stream._on_stream_event_out = Mock()
        stream._on_message_out = Mock()
        assert await stream.ack(event)
        stream._on_stream_event_out.assert_called_once_with(
            event.message.tp,
            event.message.offset,
            stream,
            event,
        )
        stream._on_message_out.assert_called_once_with(
            event.message.tp,
            event.message.offset,
            event.message,
        )

    @pytest.mark.asyncio
    async def test__format_key__callable_raises(self, *, stream):
        keyfun = Mock(name='keyfun')
        keyfun.side_effect = KeyboardInterrupt()

        with pytest.raises(Skip):
            await stream._format_key(keyfun, 300)



import random
from datetime import datetime, timezone
from itertools import count
import faust


class Withdrawal(faust.Record, isodates=True, serializer='json'):
    user: str
    country: str
    amount: float
    date: datetime = None


def generate_withdrawals(n: int = None):
    for d in generate_withdrawals_dict(n):
        yield Withdrawal(**d)


def generate_withdrawals_dict(n: int = None):
    num_countries = 5
    countries = [f'country_{i}' for i in range(num_countries)]
    country_dist = [0.9] + ([0.10 / num_countries] * (num_countries - 1))
    num_users = 500
    users = [f'user_{i}' for i in range(num_users)]
    for _ in range(n) if n is not None else count():
        yield {
            'user': random.choice(users),
            'amount': random.uniform(0, 25_000),
            'country': random.choices(countries, country_dist)[0],
            'date': datetime.utcnow().replace(tzinfo=timezone.utc).isoformat(),
        }



"""HTTP endpoint showing statistics from the Faust monitor."""
from collections import defaultdict
from typing import List, MutableMapping, Set
from faust import web
from faust.types.tuples import TP

__all__ = ['Assignment', 'Stats', 'blueprint']

TPMap = MutableMapping[str, List[int]]


blueprint = web.Blueprint('monitor')


@blueprint.route('/', name='index')
class Stats(web.View):
    """Monitor statistics."""

    async def get(self, request: web.Request) -> web.Response:
        """Return JSON response with sensor information."""
        return self.json(
            {f'Sensor{i}': s.asdict()
             for i, s in enumerate(self.app.sensors)})


@blueprint.route('/assignment/', name='assignment')
class Assignment(web.View):
    """Cluster assignment information."""

    @classmethod
    def _topic_grouped(cls, assignment: Set[TP]) -> TPMap:
        tps: MutableMapping[str, List[int]] = defaultdict(list)
        for tp in sorted(assignment):
            tps[tp.topic].append(tp.partition)
        return dict(tps)

    async def get(self, request: web.Request) -> web.Response:
        """Return current assignment as a JSON response."""
        assignor = self.app.assignor
        return self.json({
            'actives': self._topic_grouped(assignor.assigned_actives()),
            'standbys': self._topic_grouped(assignor.assigned_standbys()),
        })



"""Kafka protocol extensions."""
# pragma: no cover
import abc
from typing import Type
from aiokafka.protocol.struct import Struct
from aiokafka.protocol.types import Schema


class Response(Struct, metaclass=abc.ABCMeta):  # type: ignore
    """API Response."""

    @property
    @abc.abstractmethod
    def API_KEY(self) -> int:
        """Integer identifier for api request/response."""
        pass

    @property
    @abc.abstractmethod
    def API_VERSION(self) -> int:
        """Integer of api request/response version."""
        pass

    @property
    @abc.abstractmethod
    def SCHEMA(self) -> Schema:
        """Return instance of Schema() representing the response structure."""
        pass


class Request(Struct, metaclass=abc.ABCMeta):  # type: ignore
    """API Request."""

    @property
    @abc.abstractmethod
    def API_KEY(self) -> int:
        """Integer identifier for api request."""
        pass

    @property
    @abc.abstractmethod
    def API_VERSION(self) -> int:
        """Integer of api request version."""
        pass

    @property
    @abc.abstractmethod
    def SCHEMA(self) -> Schema:
        """Return instance of Schema() representing the request structure."""
        pass

    @property
    @abc.abstractmethod
    def RESPONSE_TYPE(self) -> Type[Response]:
        """Response class associated with the api request."""
        pass

    def expect_response(self) -> bool:
        """Return True if request type does not always return response."""
        return True



import random
from collections import Counter
from faust import Stream
from ...reports import checks
from ...reports.app import SimpleCheck
from ...app import create_stress_app

counter_received = 0
found_duplicates = 0
found_gaps = 0

app = create_stress_app(
    name='f-stress-dedupe',
    version=7,
    origin='t.stress.tests.forwarder',
    broker_commit_every=100,
)

app.add_system_check(
    checks.Increasing(
        'counter_received',
        get_value=lambda: counter_received,
    ),
)
app.add_system_check(
    checks.Stationary(
        'duplicates',
        get_value=lambda: found_duplicates,
    ),
)
app.add_system_check(
    checks.Stationary(
        'gaps',
        get_value=lambda: found_gaps,
    ),
)


leader_sending = SimpleCheck('leader-sending')

partitions_sent_counter = Counter()


@app.task
async def on_leader_send_monotonic_counter(app, max_latency=0.08) -> None:
    # Leader node sends incrementing numbers to a topic

    partitions_sent_counter.clear()

    while not app.should_stop:
        if app.rebalancing:
            partitions_sent_counter.clear()
            await app.sleep(5)
        if app.is_leader():
            await leader_sending.send_ok(app)
            for partition in range(app.conf.topic_partitions):
                if app.rebalancing:
                    break
                current_value = partitions_sent_counter.get(partition, 0)
                await process.send(value=current_value, partition=partition)
                partitions_sent_counter[partition] += 1
            await app.sleep(random.uniform(0, max_latency))
        else:
            await app.sleep(1)


@app.agent(value_type=int)
async def process(numbers: Stream[int]) -> None:
    # Next agent reads from topic and forwards numbers to another agent
    # We keep the partitioning
    async for event in numbers.events():
        await check.send(value=event.value, partition=event.message.partition)


@app.agent(value_type=int)
async def check(forwarded_numbers: Stream[int]) -> None:
    # last agent recveices number and verifies numbers are always increasing.
    # (repeating or decreasing numbers are evidence of duplicates).
    global counter_received
    global found_duplicates
    global found_gaps
    value_by_partition = Counter()
    async for event in forwarded_numbers.events():
        number = event.value
        assert isinstance(number, int)
        partition = event.message.partition
        assert isinstance(partition, int)
        previous_number = value_by_partition.get(partition, None)
        if previous_number is not None:
            if number > 0:
                # consider 0 as the service being restarted.

                if number <= previous_number:
                    # number should be larger than the previous number.
                    # if that's not true it means we have a duplicate.
                    app.log.error('Found duplicate number in %r: %r',
                                  event.message.tp, number)
                    found_duplicates += 1
                else:
                    if number != previous_number + 1:
                        # the number should always be exactly one up from the
                        # last, otherwise we dropped a number.
                        app.log.error(
                            'Sequence gap for tp %r: this=%r previous=%r',
                            event.message.tp, number, previous_number)
                        found_gaps += 1
        value_by_partition[partition] = number
        counter_received += 1



import base64
from typing import Mapping
from faust.serializers.codecs import (
    Codec, binary as _binary, codecs, dumps, get_codec, json, loads, register,
)
from faust.exceptions import ImproperlyConfigured
from faust.utils import json as _json
from hypothesis import given
from hypothesis.strategies import binary, dictionaries, text
from mode.utils.compat import want_str
import pytest
from mode.utils.mocks import patch

DATA = {'a': 1, 'b': 'string'}


def test_interface():
    s = Codec()
    with pytest.raises(NotImplementedError):
        s._loads(b'foo')
    with pytest.raises(NotImplementedError):
        s.dumps(10)
    assert s.__or__(1) is NotImplemented


@pytest.mark.parametrize('codec', ['json', 'pickle', 'yaml'])
def test_json_subset(codec: str) -> None:
    assert loads(codec, dumps(codec, DATA)) == DATA


def test_missing_yaml_library() -> None:
    msg = 'Missing yaml: pip install PyYAML'

    with patch('faust.serializers.codecs._yaml', None):
        with pytest.raises(ImproperlyConfigured):
            loads('yaml', dumps('yaml', DATA))
            pytest.fail(msg)

        with pytest.raises(ImproperlyConfigured):
            get_codec('yaml').loads(b'')
            pytest.fail(msg)


@given(binary())
def test_binary(input: bytes) -> None:
    assert loads('binary', dumps('binary', input)) == input


@given(dictionaries(text(), text()))
def test_combinators(input: Mapping[str, str]) -> None:
    s = json() | _binary()
    assert repr(s).replace("u'", "'") == 'json() | binary()'

    d = s.dumps(input)
    assert isinstance(d, bytes)
    assert _json.loads(want_str(base64.b64decode(d))) == input


def test_get_codec():
    assert get_codec('json|binary')
    assert get_codec(Codec) is Codec


def test_register():
    try:
        class MyCodec(Codec):
            ...
        register('mine', MyCodec)
        assert get_codec('mine') is MyCodec
    finally:
        codecs.pop('mine')


def test_raw():
    bits = get_codec('raw').dumps('foo')
    assert isinstance(bits, bytes)
    assert get_codec('raw').loads(bits) == b'foo'



"""Cache interface."""
import hashlib

from contextlib import suppress
from functools import wraps
from typing import (
    Any,
    Callable,
    ClassVar,
    Mapping,
    Optional,
    Type,
    Union,
    cast,
)
from urllib.parse import quote

from mode.utils.compat import want_bytes
from mode.utils.times import Seconds, want_seconds
from mode.utils.logging import get_logger

from faust.types.web import (
    CacheBackendT,
    CacheT,
    Request,
    Response,
    View,
)

logger = get_logger(__name__)

IDENT: str = 'faustweb.cache.view'


class Cache(CacheT):
    """Cache interface."""

    ident: ClassVar[str] = IDENT

    def __init__(self,
                 timeout: Seconds = None,
                 include_headers: bool = False,
                 key_prefix: str = None,
                 backend: Union[Type[CacheBackendT], str] = None,
                 **kwargs: Any) -> None:
        self.timeout = timeout
        self.include_headers = include_headers
        self.key_prefix = key_prefix or ''
        self.backend = backend

    def view(self,
             timeout: Seconds = None,
             include_headers: bool = False,
             key_prefix: str = None,
             **kwargs: Any) -> Callable[[Callable], Callable]:
        """Decorate view to be cached."""

        def _inner(fun: Callable) -> Callable:
            @wraps(fun)
            async def cached(view: View, request: Request, *args: Any,
                             **kwargs: Any) -> Response:
                key: Optional[str] = None
                is_head = request.method.upper() == 'HEAD'
                if self.can_cache_request(request):
                    key = self.key_for_request(request, key_prefix, 'GET',
                                               include_headers)

                    response = await self.get_view(key, view)
                    if response is not None:
                        logger.info('Found cached response for %r', key)
                        return response
                    if is_head:
                        response = await self.get_view(
                            self.key_for_request(request, key_prefix, 'HEAD',
                                                 include_headers),
                            view,
                        )
                        if response is not None:
                            logger.info('Found cached HEAD response for %r',
                                        key)
                            return response

                logger.info('No cache found for %r', key)
                res = await fun(view, request, *args, **kwargs)

                if key is not None and self.can_cache_response(request, res):
                    logger.info('Saving cache for key %r', key)
                    if is_head:
                        key = self.key_for_request(request, key_prefix, 'HEAD',
                                                   include_headers)
                    await self.set_view(key, view, res, timeout)
                return res

            return cached

        return _inner

    async def get_view(self,
                       key: str, view: View) -> Optional[Response]:
        """Get cached value for HTTP view request."""
        backend = self._view_backend(view)
        with suppress(backend.Unavailable):
            payload = await backend.get(key)
            if payload is not None:
                return view.bytes_to_response(payload)
        return None

    def _view_backend(self, view: View) -> CacheBackendT:
        return cast(CacheBackendT, self.backend or view.app.cache)

    async def set_view(self, key: str, view: View, response: Response,
                       timeout: Seconds = None) -> None:
        """Set cached value for HTTP view request."""
        backend = self._view_backend(view)
        _timeout = timeout if timeout is not None else self.timeout
        with suppress(backend.Unavailable):
            return await backend.set(
                key,
                view.response_to_bytes(response),
                want_seconds(_timeout) if _timeout is not None else None,
            )

    def can_cache_request(self, request: Request) -> bool:
        """Return :const:`True` if we can cache this type of HTTP request."""
        return True

    def can_cache_response(self, request: Request, response: Response) -> bool:
        """Return :const:`True` for HTTP status codes we CAN cache."""
        return response.status == 200

    def key_for_request(self,
                        request: Request,
                        prefix: str = None,
                        method: str = None,
                        include_headers: bool = False) -> str:
        """Return a cache key created from web request."""
        actual_method: str = request.method if method is None else method
        headers = request.headers if include_headers else {}
        if prefix is None:
            prefix = self.key_prefix
        return self.build_key(request, actual_method, prefix, headers)

    def build_key(self, request: Request, method: str, prefix: str,
                  headers: Mapping[str, str]) -> str:
        """Build cache key from web request and environment."""
        context = hashlib.md5(b''.join(
            want_bytes(k) + want_bytes(v)
            for k, v in headers.items())).hexdigest()
        url = hashlib.md5(iri_to_uri(str(
            request.url)).encode('ascii')).hexdigest()
        return f'{self.ident}.{prefix}.{method}.{url}.{context}'


def iri_to_uri(iri: str) -> str:
    """Convert IRI to URI."""
    # The list of safe characters here is constructed from the "reserved" and
    # "unreserved" characters specified in sections 2.2 and 2.3 of RFC 3986:
    #     reserved    = gen-delims / sub-delims
    #     gen-delims  = ":" / "/" / "?" / "#" / "[" / "]" / "@"
    #     sub-delims  = "!" / "$" / "&" / "'" / "(" / ")"
    #                   / "*" / "+" / "," / ";" / "="
    #     unreserved  = ALPHA / DIGIT / "-" / "." / "_" / "~"
    # Of the unreserved characters, urllib.parse.quote() already considers all
    # but the ~ safe.
    # The % character is also added to the list of safe characters here, as the
    # end of section 3.1 of RFC 3987 specifically mentions that % must not be
    # converted.
    return quote(iri, safe='''/#%[]=:;$&()+,!?*@'~''')



import asyncio
from copy import copy

import faust
import pytest
from faust.exceptions import ImproperlyConfigured
from faust.streams import maybe_forward
from mode import label
from mode.utils.aiter import aiter, anext
from mode.utils.mocks import AsyncMock, Mock

from .helpers import channel_empty, message, put


class FooModel(faust.Record):
    foo: str


def new_stream(app, *args, **kwargs):
    app = _prepare_app(app)
    return _new_stream(app, app.channel(loop=app.loop, maxsize=1000), **kwargs)


def new_topic_stream(app, *args, name: str = 'test', **kwargs):
    app = _prepare_app(app)
    return _new_stream(app, app.topic(name, loop=app.loop), **kwargs)


def _new_stream(app, channel, *args, **kwargs):
    return channel.stream(*args, loop=app.loop, **kwargs)


def _prepare_app(app):
    loop = asyncio.get_event_loop()
    app.loop = loop
    app.flow_control.resume()  # <-- flow control initially suspended
    return app


@pytest.mark.asyncio
@pytest.mark.allow_lingering_tasks(count=1)
async def test_simple(app, loop):
    async with new_stream(app) as stream:
        stream_it = aiter(stream)
        assert await channel_empty(stream.channel)
        await put(stream.channel, key='key', value='value')
        assert await anext(stream_it) == 'value'
        assert await channel_empty(stream.channel)


@pytest.mark.asyncio
async def test_async_iterator(app):
    async with new_stream(app) as stream:
        for i in range(100):
            await stream.channel.deliver(message(key=i, value=i))
        received = 0
        async for value in stream:
            assert value == received
            received += 1
            if received >= 100:
                break
        assert await channel_empty(stream.channel)


@pytest.mark.asyncio
async def test_throw(app):
    async with new_stream(app) as stream:
        streamit = aiter(stream)
        await stream.channel.deliver(message(key='key', value='val'))
        assert await anext(streamit) == 'val'
        await stream.throw(KeyError('foo'))
        with pytest.raises(KeyError):
            await anext(streamit)


@pytest.mark.asyncio
async def test_enumerate(app):
    async with new_stream(app) as stream:
        for i in range(100):
            await stream.channel.deliver(message(key=i, value=i * 4))
        async for i, value in stream.enumerate():
            current_event = stream.current_event
            assert i == current_event.key
            assert value == i * 4
            if i >= 99:
                break
        assert await channel_empty(stream.channel)


@pytest.mark.asyncio
async def test_items(app):
    async with new_stream(app) as stream:
        for i in range(100):
            await stream.channel.deliver(message(key=i, value=i * 2))
        i = 0
        async for key, value in stream.items():
            assert key == i
            assert value == i * 2
            i += 1
            if i > 99:
                break
        assert await channel_empty(stream.channel)


@pytest.mark.asyncio
async def test_through(app):
    app._attachments.enabled = False
    async with new_stream(app) as orig:
        channel = app.channel(loop=app.loop)
        stream = orig.through(channel)
        for i in range(100):
            await orig.channel.deliver(message(key=i, value=i * 2))

        assert stream.get_root_stream() is orig
        assert orig._passive
        # noop
        orig._enable_passive(orig.channel)

        events = []
        async for i, value in stream.enumerate():
            assert value == i * 2
            assert stream.current_event
            events.append(mock_event_ack(stream.current_event))
            if i >= 99:
                break
        await orig.stop()
        await stream.stop()
        assert_events_acked(events)


def test_through_with_concurrency_index(app):
    s = new_stream(app)
    s.concurrency_index = 0

    with pytest.raises(ImproperlyConfigured):
        s.through('foo')


@pytest.mark.asyncio
async def test_through_twice(app):
    async with new_topic_stream(app) as s:
        s._enable_passive = Mock()
        async with s.through('bar') as s2:
            s2._enable_passive = Mock()
            with pytest.raises(ImproperlyConfigured):
                s.through('baz')


def test_group_by_with_concurrency_index(app):
    s = new_stream(app)
    s.concurrency_index = 0

    with pytest.raises(ImproperlyConfigured):
        s.group_by(lambda s: s.foo)


def test_group_by_callback_must_have_name(app):
    s = new_topic_stream(app)
    with pytest.raises(TypeError):
        s.group_by(lambda s: s.foo)


@pytest.mark.asyncio
async def test_group_by__with_prefix(app):
    expected_topic = 'foo.bar.baz-test-FooModel.foo-repartition'
    async with new_topic_stream(app) as stream:
        stream.prefix = 'foo.bar.baz'
        stream._enable_passive = Mock()
        async with stream.group_by(FooModel.foo) as s2:
            assert s2.channel.get_topic_name() == expected_topic


@pytest.mark.asyncio
async def test_group_by_twice(app):
    async with new_topic_stream(app) as s:
        s._enable_passive = Mock()
        async with s.group_by(lambda s: s.foo, name='foo') as s2:
            s2._enable_passive = Mock()
            with pytest.raises(ImproperlyConfigured):
                s.group_by(lambda s: s.foo, name='foo')


@pytest.mark.asyncio
async def test_stream_over_iterable(app):
    async with app.stream([0, 1, 2, 3, 4, 5]) as s:
        i = 0
        async for value in s:
            assert value == i
            i += 1


@pytest.mark.asyncio
async def test_stream_filter(app):
    i = 0
    values = [1000, 3000, 99, 5000, 3, 9999]

    def myfilter(value):
        return value > 1000

    async with app.stream(values).filter(myfilter) as s:
        async for value in s:
            i += 1
            assert value > 1000

    assert i == 3


@pytest.mark.asyncio
async def test_events(app):
    async with new_stream(app) as stream:
        for i in range(100):
            await stream.channel.deliver(message(key=i, value=i * 2))
            await stream.channel.queue.put(i)  # no associated event
        i = 0
        events = []
        async for event in stream.events():
            print(event, i)
            assert event.key == i
            assert event.value == i * 2
            events.append(mock_event_ack(event))
            i += 1
            if i > 99:
                break
    await stream.stop()  # stop twice
    await asyncio.sleep(0)  # have to sleep twice here for all events to be
    await asyncio.sleep(0.5)  # acked for some reason
    assert_events_acked(events)


def assert_events_acked(events):
    try:
        for event in events:
            if not event.ack.called:
                assert event.message.acked
                assert not event.message.refcount
    except AssertionError:
        fail_count = len([e for e in events if not e.ack.call_count])
        fail_positions = [i for i, e in enumerate(events)
                          if not e.ack.call_count]
        print(f'ACK FAILED FOR {fail_count} EVENT(S)')
        print(f'  POSITIONS: {fail_positions}')
        raise


class test_chained_streams:

    def _chain(self, app):
        root = new_stream(app)
        root._next = s1 = new_stream(app, prev=root)
        s1._next = s2 = new_stream(app, prev=s1)
        s2._next = s3 = new_stream(app, prev=s2)
        return root, s1, s2, s3

    def test_get_root_stream(self, app):
        root, s1, s2, s3 = self._chain(app)
        assert root.get_root_stream() is root
        assert s3.get_root_stream() is root
        assert s2.get_root_stream() is root
        assert s1.get_root_stream() is root

    def test_get_active_stream(self, app):
        root, s1, s2, s3 = self._chain(app)
        assert root.get_active_stream() is s3
        assert s1.get_active_stream() is s3
        assert s2.get_active_stream() is s3
        assert s3.get_active_stream() is s3

    def test_iter_ll_forwards(self, app):
        root, s1, s2, s3 = self._chain(app)
        assert list(root._iter_ll_forwards()) == [root, s1, s2, s3]
        assert list(s1._iter_ll_forwards()) == [s1, s2, s3]
        assert list(s2._iter_ll_forwards()) == [s2, s3]
        assert list(s3._iter_ll_forwards()) == [s3]
        root_root = root.get_root_stream()
        assert root_root is root
        assert list(root_root._iter_ll_forwards()) == [root, s1, s2, s3]
        s1_root = s1.get_root_stream()
        assert s1_root is root
        assert list(s1_root._iter_ll_forwards()) == [root, s1, s2, s3]
        s2_root = s2.get_root_stream()
        assert s2_root is root
        assert list(s2_root._iter_ll_forwards()) == [root, s1, s2, s3]
        s3_root = s3.get_root_stream()
        assert s3_root is root
        assert list(s3_root._iter_ll_forwards()) == [root, s1, s2, s3]

    def test_iter_ll_backwards(self, app):
        root, s1, s2, s3 = self._chain(app)
        assert list(root._iter_ll_backwards()) == [root]
        assert list(s1._iter_ll_backwards()) == [s1, root]
        assert list(s2._iter_ll_backwards()) == [s2, s1, root]
        assert list(s3._iter_ll_backwards()) == [s3, s2, s1, root]
        root_active = root.get_active_stream()
        assert root_active is s3
        assert list(root_active._iter_ll_backwards()) == [s3, s2, s1, root]
        s1_active = s1.get_active_stream()
        assert s1_active is s3
        assert list(s1_active._iter_ll_backwards()) == [s3, s2, s1, root]
        s2_active = s2.get_active_stream()
        assert s2_active is s3
        assert list(s2_active._iter_ll_backwards()) == [s3, s2, s1, root]
        s3_active = s3.get_active_stream()
        assert s3_active is s3
        assert list(s3_active._iter_ll_backwards()) == [s3, s2, s1, root]

    def test_get_active_stream__loop_in_linkedlist(self, app):
        root, s1, s2, s3 = self._chain(app)
        s2._next = s1
        with pytest.raises(RuntimeError):
            root.get_active_stream()

    def test_get_root_stream__loop_in_linkedlist(self, app):
        root, s1, s2, s3 = self._chain(app)
        s1._prev = s2
        with pytest.raises(RuntimeError):
            s3.get_root_stream()

    @pytest.mark.asyncio
    async def test_stop_stops_chain__root(self, app):
        root, s1, s2, s3 = self._chain(app)
        await self.assert_was_stopped(root, [s1, s2, s3])

    @pytest.mark.asyncio
    async def test_stop_stops_chain__s1(self, app):
        root, s1, s2, s3 = self._chain(app)
        await self.assert_was_stopped(s1, [root, s2, s3])

    @pytest.mark.asyncio
    async def test_stop_stops_chain__s2(self, app):
        root, s1, s2, s3 = self._chain(app)
        await self.assert_was_stopped(s2, [root, s1, s3])

    @pytest.mark.asyncio
    async def test_stop_stops_chain__s3(self, app):
        root, s1, s2, s3 = self._chain(app)
        await self.assert_was_stopped(s3, [root, s1, s2])

    async def assert_was_stopped(self, leader, followers):
        all_nodes = [leader] + list(followers)
        for node in all_nodes:
            assert not node._stopped.set()
        for node in all_nodes:
            node._started.set()
        await leader.stop()
        for node in all_nodes:
            assert node._stopped.is_set()


@pytest.mark.asyncio
async def test_start_and_stop_Stream(app):
    s = new_topic_stream(app)
    await _start_stop_stream(s)
    await app.producer.stop()


async def _start_stop_stream(stream):
    assert not stream._passive
    await stream.start()
    stream.__aiter__()
    assert stream.app.topics
    await stream.stop()


@pytest.mark.asyncio
async def test_ack(app):
    async with new_stream(app) as s:
        assert s.get_active_stream() is s
        await s.channel.send(value=1)
        await s.channel.send(value=2)
        event = None
        i = 1
        async for value in s:
            assert value == i
            i += 1
            last_to_ack = value == 2
            event = mock_stream_event_ack(s, return_value=last_to_ack)
            if value == 2:
                break
        assert event
        # need one sleep on Python 3.6.0-3.6.6 + 3.7.0
        # need two sleeps on Python 3.6.7 + 3.7.1 :-/
        await asyncio.sleep(0)  # needed for some reason
        await asyncio.sleep(0)  # needed for some reason
        if not event.ack.called:
            assert event.message.acked
            assert not event.message.refcount


@pytest.mark.asyncio
async def test_noack(app):
    async with new_stream(app) as s:
        new_s = s.noack()
        assert new_s is not s
        assert s.get_active_stream() is new_s
        await new_s.channel.send(value=1)
        event = None
        async for value in new_s:
            assert value == 1
            event = mock_stream_event_ack(new_s)
            break
        assert event
        # need one sleep on Python 3.6.0-3.6.6 + 3.7.0
        # need two sleeps on Python 3.6.7 + 3.7.1 :-/
        await asyncio.sleep(0)  # needed for some reason
        await asyncio.sleep(0)  # needed for some reason
        event.ack.assert_not_called()


@pytest.mark.asyncio
async def test_acked_when_raising(app):
    async with new_stream(app) as s:
        await s.channel.send(value=1)
        await s.channel.send(value=2)

        event1 = None
        with pytest.raises(RuntimeError):
            async for value in s:
                event1 = mock_stream_event_ack(s)
                assert value == 1
                raise RuntimeError
        assert event1
        # need one sleep on Python 3.6.0-3.6.6 + 3.7.0
        # need two sleeps on Python 3.6.7 + 3.7.1 :-/
        await asyncio.sleep(0)  # needed for some reason
        await asyncio.sleep(0)  # needed for some reason
        if not event1.ack.called:
            assert event1.message.acked
            assert not event1.message.refcount

        event2 = None
        with pytest.raises(RuntimeError):
            async for value in s:
                event2 = mock_stream_event_ack(s)
                assert value == 2
                raise RuntimeError
        assert event2
        # need one sleep on Python 3.6.0-3.6.6 + 3.7.0
        # need two sleeps on Python 3.6.7 + 3.7.1 :-/
        await asyncio.sleep(0)  # needed for some reason
        await asyncio.sleep(0)  # needed for some reason
        if not event2.ack.called:
            assert event2.message.acked
            assert not event2.message.refcount


@pytest.mark.asyncio
@pytest.mark.allow_lingering_tasks(count=1)
async def test_maybe_forward__when_event(app):
    async with new_stream(app) as s:
        event = await get_event_from_value(s, 'foo')
        s.channel.send = Mock(name='channel.send')
        event.forward = AsyncMock(name='event.forward')
        await maybe_forward(event, s.channel)
        event.forward.assert_called_once_with(s.channel)
        s.channel.send.assert_not_called()


@pytest.mark.asyncio
async def test_maybe_forward__when_concrete_value(app):
    s = new_stream(app)
    s.channel.send = AsyncMock(name='channel.send')
    await maybe_forward('foo', s.channel)
    s.channel.send.assert_called_once_with(value='foo')


def mock_stream_event_ack(stream, return_value=False):
    return mock_event_ack(stream.current_event, return_value=return_value)


def mock_event_ack(event, return_value=False):
    event.ack = Mock(name='ack')
    event.ack.return_value = return_value
    return event


async def get_event_from_value(stream, value, key=None):
    await stream.channel.send(key=key, value=value)
    async for value in stream:
        event = stream.current_event
        assert event
        event.ack = Mock(name='event.ack')
        return event


def test_copy(app):
    s = new_stream(app)
    s2 = copy(s)
    assert s2 is not s
    assert s2.channel is s.channel


def test_repr(app):
    assert repr(new_stream(app))


def test_repr__combined(app):
    assert repr(new_stream(app) & new_stream(app))


def test_label(app):
    assert label(new_stream(app))


def test_iter_raises(app):
    with pytest.raises(NotImplementedError):
        for _ in new_stream(app):
            assert False


def test_derive_topic_from_nontopic_channel_raises(app):
    with pytest.raises(ValueError):
        new_stream(app).derive_topic('bar')


@pytest.mark.asyncio
async def test_remove_from_stream(app):
    s = new_stream(app)
    await s.start()
    assert not s.should_stop
    await s.remove_from_stream(s)
    assert s.should_stop


@pytest.mark.asyncio
async def test_stop_stops_related_streams(app):
    s1 = new_stream(app)
    s2 = new_stream(app)
    await s1.start()
    await s2.start()
    s3 = s1 & s2
    await s3.stop()
    assert s1.should_stop
    assert s2.should_stop
    assert s3.should_stop


@pytest.mark.asyncio
async def test_take(app):
    async with new_stream(app) as s:
        assert s.enable_acks is True
        await s.channel.send(value=1)
        event = None
        async for value in s.take(1, within=1):
            assert value == [1]
            assert s.enable_acks is False
            event = mock_stream_event_ack(s)
            break

        assert event
        # need one sleep on Python 3.6.0-3.6.6 + 3.7.0
        # need two sleeps on Python 3.6.7 + 3.7.1 :-/
        await asyncio.sleep(0)  # needed for some reason
        await asyncio.sleep(0)  # needed for some reason

        if not event.ack.called:
            assert event.message.acked
            assert not event.message.refcount
        assert s.enable_acks is True


@pytest.mark.asyncio
async def test_take__10(app, loop):
    s = new_stream(app)
    async with s:
        assert s.enable_acks is True
        for i in range(9):
            await s.channel.send(value=i)

        async def in_one_second_finalize():
            await s.sleep(1.0)
            await s.channel.send(value=9)
            for i in range(10):
                await s.channel.send(value=i + 10)
        asyncio.ensure_future(in_one_second_finalize())

        event = None
        buffer_processor = s.take(10, within=10.0)
        async for value in buffer_processor:
            assert value == list(range(10))
            assert s.enable_acks is False
            event = mock_stream_event_ack(s)
            break
        async for value in buffer_processor:
            assert value == list(range(10, 20))
            assert s.enable_acks is False
            break

        try:
            await buffer_processor.athrow(asyncio.CancelledError())
        except asyncio.CancelledError:
            pass

        assert event
        # need one sleep on Python 3.6.0-3.6.6 + 3.7.0
        # need two sleeps on Python 3.6.7 + 3.7.1 :-/
        await asyncio.sleep(0)  # needed for some reason
        await asyncio.sleep(0)  # needed for some reason
        await asyncio.sleep(0)  # needed for some reason

    if not event.ack.called:
        assert event.message.acked
        assert not event.message.refcount
    assert s.enable_acks is True


@pytest.mark.asyncio
async def test_take__no_event_crashes(app, loop):

    class NoCurrentEventStream(faust.Stream):

        @property
        def current_event(self):
            return None

        @current_event.setter
        def current_event(self, event):
            pass
    app.conf.Stream = NoCurrentEventStream

    s = new_stream(app)
    assert isinstance(s, NoCurrentEventStream)
    async with s:
        assert s.enable_acks is True
        await s.channel.send(value=1)
        buffer_processor = s.take(10, within=10.0)
        print('STARTING STREAM ITERATION')
        async for value in buffer_processor:
            print(f'RECEIVED VALUE: {value!r}')
            break
        print('ENDING STREAM ITERATION')

        try:
            await buffer_processor.athrow(asyncio.CancelledError())
        except asyncio.CancelledError:
            pass

        # need one sleep on Python 3.6.0-3.6.6 + 3.7.0
        # need two sleeps on Python 3.6.7 + 3.7.1 :-/
        await asyncio.sleep(0)  # needed for some reason
        await asyncio.sleep(0)  # needed for some reason
        await asyncio.sleep(0)  # needed for some reason
        assert isinstance(s._crash_reason, RuntimeError)
    print('RETURNING')
    assert s.enable_acks is True



import random
from collections import Counter
from copy import copy
from faust import Stream
from ...reports import checks
from ...app import create_stress_app

keys_set = 0
found_duplicates = 0
found_gaps = 0

app = create_stress_app(
    name='f-stress-tables',
    version=7,
    origin='t.stress.tests.tables',
    broker_commit_every=100,
)

processed_total = 0
processed_by_partition = Counter()


class PartitionsNotStarved(checks.Check):
    description = 'fine'
    negate_description = 'starved'

    def compare(self, prev_value, current_value):
        # takes two collections.Counter as argument
        assert prev_value is not current_value

        # yeah, this is a hack, but we don't want to have 100 threads
        # checking this, and no time to make this better as of now. [ask]
        differing = {}
        for partition, prev_number in prev_value.items():
            current_number = current_value[partition]
            if current_value[partition] <= prev_number:
                differing[partition] = (prev_number, current_number)

        if differing:
            self.prev_value_repr = self.current_value_repr
            self.current_value_repr = repr(differing)
            if len(differing) == len(current_value):
                self.current_value_repr = 'all partitions starved'

        return bool(differing)

    def get_value(self):
        return processed_by_partition

    def store_previous_value(self, current_value):
        self.prev_value = copy(current_value)


app.add_system_check(
    PartitionsNotStarved('starved-partitions'),
)
app.add_system_check(
    checks.Stationary(
        'table-found-double-count',
        get_value=lambda: found_duplicates,
    ),
)
app.add_system_check(
    checks.Stationary(
        'table-found-missing-update',
        get_value=lambda: found_gaps,
    ),
)

partitions_sent_counter = Counter()

table = app.Table('counted', key_type=int, value_type=int)


@app.task
async def on_leader_send_monotonic_counter(app, max_latency=0.01) -> None:
    # Leader node sends incrementing numbers to a topic

    partitions_sent_counter.clear()

    while not app.should_stop:
        if app.rebalancing:
            partitions_sent_counter.clear()
            await app.sleep(5)
        if app.is_leader():
            for partition in range(app.conf.topic_partitions):
                if app.rebalancing:
                    break
                current_value = partitions_sent_counter.get(partition, 0)
                await process.send(value=current_value, partition=partition)
                partitions_sent_counter[partition] += 1
            await app.sleep(random.uniform(0, max_latency))
        else:
            await app.sleep(1)


@app.agent(value_type=int)
async def process(numbers: Stream[int]) -> None:
    # Next agent reads from topic and forwards numbers to another agent
    # We keep the partitioning
    global keys_set
    global found_duplicates
    global found_gaps
    global processed_total

    # This is a stream of numbers where every partition count from 1 to
    # infinity.

    # If we have processed this stream correctly
    #   - there shouldn't be any duplicate numbers
    #   - there shouldn't be any missing numbers.

    async for event in numbers.events():
        number = event.value
        assert isinstance(number, int)
        partition = event.message.partition
        previous_number = table.get(partition)
        if processed_total and not processed_total % 30_000:
            with app.system_checks.pause():
                print('Pausing stream to fill topics')
                await app.sleep(100.0)

        if previous_number is not None:
            if number > 0:
                # 0 is fine, means leader restarted.

                if number <= previous_number:
                    # Remember, the key is the partition.
                    # So the previous value set for any partition,
                    # should always be the same as the previous value in the
                    # topic.
                    #
                    # if number is less than previous number, it means
                    # we have missed counting one (the sequence counts
                    # WITH THE OFFSET after all!)
                    # if the number is less than we have a problem.
                    app.log.error('Found duplicate number in %r: %r',
                                  event.message.tp, number)
                    found_duplicates += 1
                else:
                    if number != previous_number + 1:
                        # number should always be one up from the last,
                        # otherwise we dropped a number.
                        app.log.error(
                            'Sequence gap for tp %r: this=%r previous=%r',
                            event.message.tp, number, previous_number)
                        found_gaps += 1
        table[partition] = number
        processed_by_partition[partition] += 1
        processed_total += 1



#!/usr/bin/env python
from typing import AsyncIterable
import faust
from faust import StreamT
from faust.cli import argument


app = faust.App('RPC99', reply_create_topic=True)
pow_topic = app.topic('RPC__pow')
mul_topic = app.topic('RPC__mul')


@app.agent(pow_topic)
async def pow(stream: StreamT[float]) -> AsyncIterable[float]:
    async for value in stream:
        yield await mul.ask(value=value ** 2)


@app.agent(mul_topic)
async def mul(stream: StreamT[float]) -> AsyncIterable[float]:
    async for value in stream:
        yield value * 100.0


@app.timer(interval=10.0)
async def _sender() -> None:
    # join' gives list with order preserved.
    res = await pow.join([30.3, 40.4, 50.5, 60.6, 70.7, 80.8, 90.9])
    print(f'JOINED: {res!r}')

    # map' gives async iterator to stream results (unordered)
    #   note: the argument can also be an async iterator.
    async for value in pow.map([30.3, 40.4, 50.5, 60.6, 70.7, 80.8, 90.9]):
        print(f'RECEIVED REPLY: {value!r}')


@app.command(argument('x'))
async def x100(self, x):
    res = await mul.ask(float(x))
    print(f'{x} * 100 = {res}')


if __name__ == '__main__':
    app.main()



from ...app import create_stress_app
from ...models import Withdrawal, generate_withdrawals

app = create_stress_app(
    name='f-stress-simple',
    version=7,
    origin='t.stress.tests.simple',
)

withdrawals_topic = app.topic('f-stress-withdrawals', value_type=Withdrawal)


@app.register_stress_producer
async def produce_withdrawals(max_messages):
    for withdrawal in generate_withdrawals(max_messages):
        yield await withdrawals_topic.send(key=None, value=withdrawal)


@app.agent(withdrawals_topic)
async def track_user_withdrawal(withdrawals):
    async for _withdrawal in withdrawals:  # noqa
        app.count_received_events += 1



"""LiveCheck - Faust Application."""
import asyncio

from datetime import timedelta
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    Type,
    cast,
)

from mode.signals import BaseSignalT
from mode.utils.compat import want_bytes
from mode.utils.objects import annotations, cached_property, qualname
from mode.utils.times import Seconds

import faust
from faust.app.base import SCAN_CATEGORIES
from faust.sensors.base import Sensor
from faust.types import AgentT, AppT, EventT, StreamT, TP, TopicT
from faust.utils import venusian

from . import patches
from .case import Case
from .exceptions import LiveCheckError
from .locals import current_test, current_test_stack
from .models import SignalEvent, TestExecution, TestReport
from .signals import BaseSignal, Signal

__all__ = ['LiveCheck']

SCAN_CASE = 'livecheck.case'

#: alias for mypy bug
_Case = Case

patches.patch_all()  # XXX


class LiveCheckSensor(Sensor):

    def on_stream_event_in(self,
                           tp: TP,
                           offset: int,
                           stream: StreamT,
                           event: EventT) -> Optional[Dict]:
        """Call when stream starts processing event."""
        test = TestExecution.from_headers(event.headers)
        if test is not None:
            stream.current_test = test  # type: ignore
            current_test_stack.push_without_automatic_cleanup(test)
        return None

    def on_stream_event_out(self,
                            tp: TP,
                            offset: int,
                            stream: StreamT,
                            event: EventT,
                            state: Dict = None) -> None:
        """Call when stream is finished handling event."""
        has_active_test = getattr(stream, 'current_test', None)
        if has_active_test:
            stream.current_test = None  # type: ignore
            current_test_stack.pop()


class LiveCheck(faust.App):
    """LiveCheck application."""

    SCAN_CATEGORIES = list(SCAN_CATEGORIES) + [SCAN_CASE]

    Signal: ClassVar[Type[BaseSignal]]
    Signal = Signal

    Case: ClassVar[Type[_Case]]
    Case = _Case

    #: Number of concurrent actors processing signal events.
    bus_concurrency: int = 30

    #: Number of concurrent actors executing test cases.
    test_concurrency: int = 100

    #: Unset this if you don't want reports to be sent to
    #: the :attr:`report_topic_name` topic.
    send_reports: bool = True

    test_topic_name: str = 'livecheck'
    bus_topic_name: str = 'livecheck-bus'
    report_topic_name: str = 'livecheck-report'

    cases: Dict[str, _Case]

    _resolved_signals: Dict[Tuple[str, str, Any], SignalEvent]

    @classmethod
    def for_app(cls, app: AppT, *,
                prefix: str = 'livecheck-',
                web_port: int = 9999,
                test_topic_name: str = None,
                bus_topic_name: str = None,
                report_topic_name: str = None,
                bus_concurrency: int = None,
                test_concurrency: int = None,
                send_reports: bool = None,
                **kwargs: Any) -> 'LiveCheck':
        """Create LiveCheck application targeting specific app.

        The target app will be used to configure the LiveCheck app.
        """
        app_id, passed_kwargs = app._default_options
        livecheck_id = f'{prefix}{app_id}'
        override = {
            'web_port': web_port,
            'test_topic_name': test_topic_name,
            'bus_topic_name': bus_topic_name,
            'report_topic_name': report_topic_name,
            'bus_concurrency': bus_concurrency,
            'test_concurrency': test_concurrency,
            'send_reports': send_reports,
            **kwargs}
        options = {**passed_kwargs, **override}

        livecheck_app = cls(livecheck_id, **options)
        livecheck_app._contribute_to_app(app)

        return livecheck_app

    def _contribute_to_app(self, app: AppT) -> None:
        from .patches.aiohttp import LiveCheckMiddleware
        web_app = app.web.web_app  # type: ignore
        web_app.middlewares.append(LiveCheckMiddleware())
        app.sensors.add(LiveCheckSensor())
        app.livecheck = self  # type: ignore

    def __init__(self,
                 id: str,
                 *,
                 test_topic_name: str = None,
                 bus_topic_name: str = None,
                 report_topic_name: str = None,
                 bus_concurrency: int = None,
                 test_concurrency: int = None,
                 send_reports: bool = None,
                 **kwargs: Any) -> None:
        super().__init__(id, **kwargs)

        if test_topic_name is not None:
            self.test_topic_name = test_topic_name
        if bus_topic_name is not None:
            self.bus_topic_name = bus_topic_name
        if report_topic_name is not None:
            self.report_topic_name = report_topic_name
        if bus_concurrency is not None:
            self.bus_concurrency = bus_concurrency
        if test_concurrency is not None:
            self.test_concurrency = test_concurrency
        if send_reports is not None:
            self.send_reports = send_reports

        self.cases = {}
        self._resolved_signals = {}
        patches.patch_all()
        self._apply_monkeypatches()
        self._connect_signals()

    @property
    def current_test(self) -> Optional[TestExecution]:
        """Return the current test context (if any)."""
        return current_test()

    @cached_property
    def _can_resolve(self) -> asyncio.Event:
        return asyncio.Event()

    def _apply_monkeypatches(self) -> None:
        patches.patch_all()

    def _connect_signals(self) -> None:
        AppT.on_produce_message.connect(
            self.on_produce_attach_test_headers)  # type: ignore

    def on_produce_attach_test_headers(
            self,
            sender: AppT,
            key: bytes = None,
            value: bytes = None,
            partition: int = None,
            timestamp: float = None,
            headers: List[Tuple[str, bytes]] = None,
            signal: BaseSignalT = None,
            **kwargs: Any) -> None:
        """Attach test headers to Kafka produce requests."""
        test = current_test()
        if test is not None:
            if headers is None:
                raise TypeError('Produce request missing headers list')
            headers.extend([
                (k, want_bytes(v)) for k, v in test.as_headers().items()
            ])

    def case(self, *,
             name: str = None,
             probability: float = None,
             warn_stalled_after: Seconds = timedelta(minutes=30),
             active: bool = None,
             test_expires: Seconds = None,
             frequency: Seconds = None,
             max_history: int = None,
             max_consecutive_failures: int = None,
             url_timeout_total: float = None,
             url_timeout_connect: float = None,
             url_error_retries: float = None,
             url_error_delay_min: float = None,
             url_error_delay_backoff: float = None,
             url_error_delay_max: float = None,
             base: Type[_Case] = Case) -> Callable[[Type], _Case]:
        """Decorate class to be used as a test case.

        Returns:
            :class:`faust.livecheck.Case`.
        """
        base_case = base

        def _inner(cls: Type) -> _Case:
            case_cls = type(cls.__name__, (cls, base_case), {
                '__module__': cls.__module__,
                'app': self,
            })

            signal_types = dict(self._extract_signals(case_cls, base_case))
            signals = []

            for i, (attr_name, attr_type) in enumerate(signal_types.items()):
                signal = getattr(case_cls, attr_name, None)
                if signal is None:
                    signal = attr_type(name=attr_name, index=i + 1)
                    setattr(case_cls, attr_name, signal)
                    signals.append(signal)
                else:
                    signal.index = i + 1

            case = self.add_case(case_cls(
                app=self,
                name=self._prepare_case_name(name or qualname(cls)),
                active=active,
                probability=probability,
                warn_stalled_after=warn_stalled_after,
                signals=signals,
                test_expires=test_expires,
                frequency=frequency,
                max_history=max_history,
                max_consecutive_failures=max_consecutive_failures,
                url_timeout_total=url_timeout_total,
                url_timeout_connect=url_timeout_connect,
                url_error_retries=url_error_retries,
                url_error_delay_min=url_error_delay_min,
                url_error_delay_backoff=url_error_delay_backoff,
                url_error_delay_max=url_error_delay_max,
            ))
            venusian.attach(cast(Callable, case), category=SCAN_CASE)
            return case

        return _inner

    def _extract_signals(
            self,
            case_cls: Type[_Case],
            base_case: Type[_Case]) -> Iterable[Tuple[str, Type[BaseSignal]]]:
        fields, defaults = annotations(
            case_cls,
            stop=base_case,
            skip_classvar=True,
            localns={case_cls.__name__: case_cls},
        )

        for attr_name, attr_type in fields.items():
            actual_type = getattr(attr_type, '__origin__', attr_type)
            if actual_type is None:  # Python <3.7
                actual_type = attr_type
            try:
                if issubclass(actual_type, BaseSignal):
                    yield attr_name, attr_type
            except TypeError:  # pragma: no cover
                pass  # pragma: no cover

    def add_case(self, case: _Case) -> _Case:
        """Add and register new test case."""
        self.cases[case.name] = case
        return case

    async def post_report(self, report: TestReport) -> None:
        """Publish test report to reporting topic."""
        key = None
        if report.test is not None:
            key = report.test.id
        await self.reports.send(key=key, value=report)

    async def on_start(self) -> None:
        """Call when LiveCheck application starts."""
        await super().on_start()
        self._install_bus_agent()
        self._install_test_execution_agent()

    async def on_started(self) -> None:
        """Call when LiveCheck application is fully started."""
        await super().on_started()
        for case in self.cases.values():
            await self.add_runtime_dependency(case)

    def _install_bus_agent(self) -> AgentT:
        return self.agent(
            channel=self.bus,
            concurrency=self.bus_concurrency,
        )(self._populate_signals)

    def _install_test_execution_agent(self) -> AgentT:
        return self.agent(
            channel=self.pending_tests,
            concurrency=self.test_concurrency,
        )(self._execute_tests)

    async def _populate_signals(self, events: StreamT[SignalEvent]) -> None:
        async for test_id, event in events.items():
            event.case_name = self._prepare_case_name(event.case_name)
            try:
                case = self.cases[event.case_name]
            except KeyError:
                self.log.error('Received signal %r for unregistered case %r',
                               event, (test_id, event.case_name))
            else:
                await case.resolve_signal(test_id, event)

    async def _execute_tests(self, tests: StreamT[TestExecution]) -> None:
        async for test_id, test in tests.items():
            test.case_name = self._prepare_case_name(test.case_name)
            try:
                case = self.cases[test.case_name]
            except KeyError:
                self.log.error('Unregistered test case %r with id %r: %r',
                               test.case_name, test_id, test)
            else:
                try:
                    await case.execute(test)
                except LiveCheckError:
                    pass

    def _prepare_case_name(self, name: str) -> str:
        if name.startswith('__main__.'):
            if not self.conf.origin:
                raise RuntimeError('LiveCheck app missing origin argument')
            return self.conf.origin + name[8:]
        return name

    @cached_property
    def bus(self) -> TopicT:
        """Topic used for signal communication."""
        return self.topic(
            self.bus_topic_name,
            key_type=str,
            value_type=SignalEvent,
        )

    @cached_property
    def pending_tests(self) -> TopicT:
        """Topic used to keep pending test executions."""
        return self.topic(
            self.test_topic_name,
            key_type=str,
            value_type=TestExecution,
        )

    @cached_property
    def reports(self) -> TopicT:
        """Topic used to log test reports."""
        return self.topic(
            self.report_topic_name,
            key_type=str,
            value_type=TestReport,
        )



from pathlib import Path
import pytest
from faust import web
from mode.utils.mocks import Mock


class test_Blueprint:

    @pytest.fixture()
    def bp(self):
        return web.Blueprint('test')

    def test_cache__with_prefix(self, *, bp):
        assert bp.cache(key_prefix='kp_').key_prefix == 'kp_'

    def test_apply_static(self, *, bp):
        bp.static('/static/', '/opt/static/', name='statici')
        assert bp.static_routes
        route = bp.static_routes[0]
        assert route.name == 'test.statici'
        app = Mock(name='app')
        bp.register(app, url_prefix='/foo/')
        app.web.add_static.assert_called_once_with(
            '/foo/static/', Path('/opt/static/'),
        )

    def test_apply_static__already_prefixed(self, *, bp):
        bp.static('/static/', '/opt/static/', name='test.statici')
        assert bp.static_routes
        route = bp.static_routes[0]
        assert route.name == 'test.statici'

    def test_repr(self, *, bp):
        assert repr(bp)



import pytest
from faust.cli.completion import completion
from mode.utils.mocks import patch


class test_completion:

    @pytest.fixture()
    def command(self, *, context):
        return completion(context)

    @pytest.mark.asyncio
    async def test_run(self, *, command):
        with patch('faust.cli.completion.click_completion') as cc:
            await command.run()
            cc.get_code.assert_called_once_with(shell=command.shell())

    @pytest.mark.asyncio
    async def test_run__no_completion(self, *, command):
        with patch('faust.cli.completion.click_completion', None):
            with pytest.raises(command.UsageError):
                await command.run()



"""Patches - LiveCheck integration with other frameworks/libraries."""
from . import aiohttp

__all__ = ['aiohttp', 'patch_all']


def patch_all() -> None:
    """Apply all LiveCheck monkey patches."""
    aiohttp.patch_all()



from pathlib import Path
import pytest
from faust.cli.clean_versions import clean_versions
from mode.utils.mocks import Mock, call, patch


class test_clean_versions:

    @pytest.fixture()
    def command(self, *, context):
        return clean_versions(context)

    @pytest.mark.asyncio
    async def test_run(self, *, command):
        command.remove_old_versiondirs = Mock()
        await command.run()
        command.remove_old_versiondirs.assert_called_with()

    def test_remove_old_versiondirs(self, *, app, command):
        app.conf.find_old_versiondirs = Mock(return_value=[
            Path('A1'), Path('B2'), Path('C3'),
        ])
        with patch('faust.cli.clean_versions.rmtree') as rmtree:
            command.remove_old_versiondirs()
            rmtree.assert_has_calls([
                call('A1'),
                call('B2'),
                call('C3'),
            ])



import faust
from faust.cli import argument

app = faust.App(
    'table-of-sets-windowed',
    origin='examples.tableofset',
    topic_partitions=4,
    version=2,
)

table = app.SetTable(
    'people', value_type=str,
)

joining_topic = app.topic('people_joining2', key_type=str, value_type=str)
leaving_topic = app.topic('people_leaving2', key_type=str, value_type=str)


@app.agent(joining_topic)
async def join(stream):
    async for key, name in stream.items():
        print(f'- {name.capitalize()} joined {key}')
        table[key].add(name)
        print(f'TABLE CUR: {table[key]!r}')


@app.agent(leaving_topic)
async def leave(stream):
    async for key, name in stream.items():
        print(f'- {name.capitalize()} left {key}')
        table[key].discard(name)
        print(f'TABLE CUR: {table[key]!r}')


@app.command(
    argument('location'),
    argument('name'),
)
async def joining(self, location: str, name: str):
    await joining_topic.send(key=location, value=name)


@app.command(
    argument('location'),
    argument('name'),
)
async def leaving(self, location: str, name: str):
    await leaving_topic.send(key=location, value=name)


@app.timer(10.0)
async def _dump():
    print(f'TABLE NOW:\n{table.as_ansitable()}')


if __name__ == '__main__':
    app.main()



import re
import collections
import faust
from faust.agents import Agent
from faust.app.base import SCAN_AGENT, SCAN_PAGE, SCAN_TASK
from faust.assignor.leader_assignor import LeaderAssignor, LeaderAssignorT
from faust.channels import Channel, ChannelT
from faust.cli.base import AppCommand
from faust.exceptions import (
    AlreadyConfiguredWarning,
    ConsumerNotStarted,
    ImproperlyConfigured,
    SameNode,
)
from faust.fixups.base import Fixup
from faust.sensors.monitor import Monitor
from faust.serializers import codecs
from faust.transport.base import Transport
from faust.transport.conductor import Conductor
from faust.transport.consumer import Consumer, Fetcher
from faust.types import TP
from faust.types.models import ModelT
from faust.types.settings import Settings
from faust.types.tables import GlobalTableT
from faust.types.web import ResourceOptions
from mode import Service
from mode.utils.compat import want_bytes
from mode.utils.mocks import ANY, AsyncMock, Mock, call, patch
from yarl import URL
import pytest

TEST_TOPIC = 'test'
CONFIG_DICT = {
    'broker': 'kafka://foo',
    'stream_buffer_maxsize': 1,
}
CONFIG_PATH = 't.unit.app.test_base.ConfigClass'

TP1 = TP('foo', 0)
TP2 = TP('bar', 1)
TP3 = TP('baz', 2)
TP4 = TP('xuz', 3)


class ConfigClass:
    broker = 'kafka://foo'
    stream_buffer_maxsize = 1


class Key(faust.Record):
    value: int


class Value(faust.Record, serializer='json'):
    amount: float


@pytest.mark.asyncio
@pytest.mark.parametrize('key,topic_name,expected_topic,key_serializer', [
    ('key', TEST_TOPIC, TEST_TOPIC, None),
    (Key(value=10), TEST_TOPIC, TEST_TOPIC, None),
    ({'key': 'k'}, TEST_TOPIC, TEST_TOPIC, 'json'),
    (None, 'topic', 'topic', None),
    (b'key', TEST_TOPIC, TEST_TOPIC, None),
    ('key', 'topic', 'topic', None),
])
async def test_send(
        key, topic_name, expected_topic, key_serializer, app):
    topic = app.topic(topic_name)
    event = Value(amount=0.0)
    await app.send(topic, key, event, key_serializer=key_serializer)
    # do it twice so producer_started is also True
    await app.send(topic, key, event, key_serializer=key_serializer)
    expected_sender = app.producer.send
    if key is not None:
        if isinstance(key, str):
            # Default serializer is raw, and str should be serialized
            # (note that a bytes key will be left alone.
            key_serializer = 'raw'
        if isinstance(key, ModelT):
            expected_key = key.dumps(serializer='raw')
        elif key_serializer:
            expected_key = codecs.dumps(key_serializer, key)
        else:
            expected_key = want_bytes(key)
    else:
        expected_key = None
    expected_sender.assert_called_with(
        expected_topic, expected_key, event.dumps(),
        partition=None,
        timestamp=None,
        headers={},
    )


@pytest.mark.asyncio
async def test_send_str(app):
    await app.send('foo', Value(amount=0.0))


class test_App:

    def test_stream(self, *, app):
        s = app.topic(TEST_TOPIC).stream()
        assert s.channel.topics == (TEST_TOPIC,)
        assert s.channel in app.topics
        assert s.channel.app == app

    def test_new_producer(self, *, app):
        app._producer = None
        transport = app._transport = Mock(
            name='transport',
            autospec=Transport,
        )
        assert app._new_producer() is transport.create_producer.return_value
        transport.create_producer.assert_called_with(beacon=ANY)
        assert app.producer is transport.create_producer.return_value

    @pytest.mark.parametrize('broker_url,broker_consumer_url', [
        ('moo://', None),
        ('moo://', 'zoo://'),
    ])
    def test_new_transport(self, broker_url, broker_consumer_url, *,
                           app, patching):
        app.conf.broker = broker_url
        if broker_consumer_url:
            app.conf.broker_consumer = broker_consumer_url
        by_url = patching('faust.transport.by_url')
        assert app._new_transport() is by_url.return_value.return_value
        assert app.transport is by_url.return_value.return_value
        by_url.assert_called_with(app.conf.broker_consumer[0])
        by_url.return_value.assert_called_with(
            app.conf.broker_consumer, app, loop=app.loop)
        app.transport = 10
        assert app.transport == 10

    @pytest.mark.parametrize('broker_url,broker_producer_url', [
        ('moo://', None),
        ('moo://', 'zoo://'),
    ])
    def test_new_producer_transport(self, broker_url, broker_producer_url, *,
                                    app, patching):
        app.conf.broker = broker_url
        if broker_producer_url:
            app.conf.broker_producer = broker_producer_url
        by_url = patching('faust.transport.by_url')
        transport = app._new_producer_transport()
        assert transport is by_url.return_value.return_value
        assert app.producer_transport is by_url.return_value.return_value
        by_url.assert_called_with(app.conf.broker_producer[0])
        by_url.return_value.assert_called_with(
            app.conf.broker_producer, app, loop=app.loop)
        app.producer_transport = 10
        assert app.producer_transport == 10

    @pytest.mark.asyncio
    async def test_on_stop(self, *, app):
        app._http_client = Mock(name='http_client', close=AsyncMock())
        app._producer = Mock(name='producer', flush=AsyncMock())
        await app.on_stop()
        app._http_client.close.assert_called_once_with()
        app._http_client = None
        await app.on_stop()
        app._producer = None
        await app.on_stop()

    @pytest.mark.asyncio
    async def test_stop_consumer__wait_empty_enabled(self, *, app):
        app.conf.stream_wait_empty = True
        await self.assert_stop_consumer(app)
        app._consumer.wait_empty.assert_called_once_with()

    @pytest.mark.asyncio
    async def test_stop_consumer__wait_empty_disabled(self, *, app):
        app.conf.stream_wait_empty = False
        await self.assert_stop_consumer(app)
        app.consumer.wait_empty.assert_not_called()

    async def assert_stop_consumer(self, app):
        consumer = app._consumer = Mock(
            wait_empty=AsyncMock(),
        )
        consumer.assignment.return_value = set()
        app.tables = Mock()
        app.flow_control = Mock()
        app._stop_fetcher = AsyncMock()
        await app._stop_consumer()

        consumer.assignment.side_effect = ConsumerNotStarted()
        await app._stop_consumer()

        consumer.assignment.side_effect = None
        assigned = {TP('foo', 0), TP('bar', 1)}
        consumer.assignment.return_value = assigned

        await app._stop_consumer()
        app.tables.on_partitions_revoked.assert_called_once_with(assigned)
        consumer.stop_flow.assert_called_once_with()
        app.flow_control.suspend.assert_called_once_with()
        app._stop_fetcher.assert_called_once_with()

    def test_on_rebalance_start__existing_state(self, *, app):
        app._rebalancing_sensor_state = {'foo': 'bar'}
        app.log.warning = Mock()
        app.on_rebalance_start()
        app.log.warning.assert_called_once()

    def test_on_rebalance_return__no_state(self, *, app):
        app._rebalancing_sensor_state = None
        app.log.warning = Mock()
        app.on_rebalance_return()
        app.log.warning.assert_called_once()

    def test_on_rebalance_return__has_state(self, *, app):
        app._rebalancing_sensor_state = {'time_start': 100.0}
        app.sensors.on_rebalance_return = Mock()
        app.on_rebalance_return()
        app.sensors.on_rebalance_return.assert_called_once_with(
            app, app._rebalancing_sensor_state,
        )

    def test_on_rebalance_start_end(self, *, app):
        app.tables = Mock()
        app.sensors = Mock()
        app.sensors.on_rebalance_start.return_value = {
            'time_start': 100.,
        }
        assert not app.rebalancing

        app.on_rebalance_start()
        assert app._rebalancing_sensor_state
        assert app.rebalancing
        app.tables.on_rebalance_start.assert_called_once_with()

        app.on_rebalance_return()

        app.on_rebalance_end()
        assert not app.rebalancing
        assert not app._rebalancing_sensor_state

        app.tracer = Mock(name='tracer')
        app.on_rebalance_start()
        span = app._rebalancing_span
        assert span is not None
        app._rebalancing_sensor_state = None
        app.log.warning = Mock()
        app.on_rebalance_end()
        span.finish.assert_called_once_with()
        app.log.warning.assert_called_once()

        assert not app.rebalancing

    def test_trace(self, *, app):
        app.tracer = None
        with app.trace('foo'):
            pass
        app.tracer = Mock()
        assert app.trace('foo') is app.tracer.trace.return_value

    def test_traced(self, *, app):
        @app.traced
        def foo(val):
            return val
        assert foo(42) == 42

    def test__start_span_from_rebalancing(self, *, app):
        app.tracer = None
        app._rebalancing_span = None
        assert app._start_span_from_rebalancing('foo')
        app.tracer = Mock(name='tracer')
        try:
            app._rebalancing_span = Mock(name='span')
            assert app._start_span_from_rebalancing('foo')
        finally:
            app.tracer = None
            app._rebalancing_span = None

    @pytest.mark.asyncio
    async def test_on_partitions_revoked(self, *, app):
        app.on_partitions_revoked = Mock(send=AsyncMock())
        consumer = app.consumer = Mock(
            wait_empty=AsyncMock(),
            transactions=Mock(
                on_partitions_revoked=AsyncMock(),
            ),
        )
        app.tables = Mock()
        app.flow_control = Mock()
        app._producer = Mock(flush=AsyncMock())
        revoked = {TP('foo', 0), TP('bar', 1)}
        ass = app.consumer.assignment.return_value = {TP('foo', 0)}

        app.in_transaction = False
        await app._on_partitions_revoked(revoked)

        app.on_partitions_revoked.send.assert_called_once_with(revoked)
        consumer.stop_flow.assert_called_once_with()
        app.flow_control.suspend.assert_called_once_with()
        consumer.pause_partitions.assert_called_once_with(ass)
        app.flow_control.clear.assert_called_once_with()
        consumer.wait_empty.assert_called_once_with()
        app._producer.flush.assert_called_once_with()
        consumer.transactions.on_partitions_revoked.assert_not_called()

        app.in_transaction = True
        await app._on_partitions_revoked(revoked)
        consumer.transactions.on_partitions_revoked.assert_called_once_with(
            revoked)

    @pytest.mark.asyncio
    async def test_on_partitions_revoked__no_assignment(self, *, app):
        app.on_partitions_revoked = Mock(send=AsyncMock())
        app.consumer = Mock()
        app.tables = Mock()
        revoked = {TP('foo', 0), TP('bar', 1)}
        app.consumer.assignment.return_value = set()
        await app._on_partitions_revoked(revoked)

        app.on_partitions_revoked.send.assert_called_once_with(revoked)

    @pytest.mark.asyncio
    async def test_on_partitions_revoked__crashes(self, *, app):
        app.on_partitions_revoked = Mock(send=AsyncMock())
        app.crash = AsyncMock()
        app.consumer = Mock()
        app.tables = Mock()
        revoked = {TP('foo', 0), TP('bar', 1)}
        app.consumer.assignment.side_effect = RuntimeError()
        await app._on_partitions_revoked(revoked)

        app.crash.assert_called_once()

    @pytest.mark.asyncio
    async def test_on_partitions_revoked__when_stopped(self, *, app):
        app._stopped.set()
        app._on_rebalance_when_stopped = AsyncMock()
        await app._on_partitions_revoked(set())
        app._on_rebalance_when_stopped.assert_called_once_with()

    @pytest.mark.asyncio
    async def test_stop_fetcher(self, *, app):
        app._fetcher = Mock(stop=AsyncMock())
        await app._stop_fetcher()
        app._fetcher.stop.assert_called_once_with()
        app._fetcher.service_reset.assert_called_once_with()

    def test_on_rebalance_when_stopped(self, *, app):
        app.consumer = Mock()
        app._on_rebalance_when_stopped()
        app.consumer.close.assert_called_once_with()

    @pytest.mark.asyncio
    async def test_on_partitions_assigned__when_stopped(self, *, app):
        app._stopped.set()
        app._on_rebalance_when_stopped = AsyncMock()
        await app._on_partitions_assigned(set())
        app._on_rebalance_when_stopped.assert_called_once_with()

    @pytest.mark.asyncio
    async def test_on_partitions_assigned(self, *, app):
        app._assignment = {TP('foo', 1), TP('bar', 2)}
        app.on_partitions_assigned = Mock(send=AsyncMock())
        app.consumer = Mock(
            transactions=Mock(
                on_rebalance=AsyncMock(),
            ),
        )
        app.agents = Mock(
            on_rebalance=AsyncMock(),
        )
        app.tables = Mock(
            on_rebalance=AsyncMock(),
        )
        app.topics = Mock(
            maybe_wait_for_subscriptions=AsyncMock(),
            on_partitions_assigned=AsyncMock(),
        )

        assigned = {TP('foo', 1), TP('baz', 3)}
        revoked = {TP('bar', 2)}
        newly_assigned = {TP('baz', 3)}

        app.in_transaction = False
        await app._on_partitions_assigned(assigned)

        app.agents.on_rebalance.assert_called_once_with(
            revoked, newly_assigned)
        app.topics.maybe_wait_for_subscriptions.assert_called_once_with()
        app.consumer.pause_partitions.assert_called_once_with(assigned)
        app.topics.on_partitions_assigned.assert_called_once_with(assigned)
        app.consumer.transactions.on_rebalance.assert_not_called()
        app.tables.on_rebalance.assert_called_once_with(
            assigned, revoked, newly_assigned)
        app.on_partitions_assigned.send.assert_called_once_with(assigned)

        app.in_transaction = True
        app._assignment = {TP('foo', 1), TP('bar', 2)}
        await app._on_partitions_assigned(assigned)
        app.consumer.transactions.on_rebalance.assert_called_once_with(
            assigned, revoked, newly_assigned)

    @pytest.mark.asyncio
    async def test_on_partitions_assigned__crashes(self, *, app):
        app._assignment = {TP('foo', 1), TP('bar', 2)}
        app.on_partitions_assigned = Mock(send=AsyncMock())
        app.consumer = Mock()
        app.agents = Mock(
            on_rebalance=AsyncMock(),
        )
        app.agents.on_rebalance.coro.side_effect = RuntimeError()
        app.crash = AsyncMock()

        await app._on_partitions_assigned(set())
        app.crash.assert_called_once()

    @pytest.mark.parametrize('prev,new,expected_revoked,expected_assigned', [
        (None, {TP1, TP2}, set(), {TP1, TP2}),
        (set(), set(), set(), set()),
        (set(), {TP1, TP2}, set(), {TP1, TP2}),
        ({TP1, TP2}, {TP1, TP2}, set(), set()),
        ({TP1, TP2}, {TP1, TP3, TP4}, {TP2}, {TP3, TP4}),
    ])
    def test_update_assignment(
            self, prev, new, expected_revoked, expected_assigned,
            *, app):
        app._assignment = prev
        revoked, newly_assigned = app._update_assignment(new)
        assert revoked == expected_revoked
        assert newly_assigned == expected_assigned
        assert app._assignment == new

    def test_worker_init(self, *, app):
        fixup1 = Mock(name='fixup1', autospec=Fixup)
        fixup2 = Mock(name='fixup2', autospec=Fixup)
        app.fixups = [fixup1, fixup2]
        app.worker_init()
        app.worker_init_post_autodiscover()
        fixup1.on_worker_init.assert_called_once_with()
        fixup2.on_worker_init.assert_called_once_with()

    def test_worker_init_post_autodiscover(self, *, app):
        on_worker_init = app.on_worker_init.connect(
            Mock(name='on_worker_init'))
        app.worker_init_post_autodiscover()
        on_worker_init.assert_called_once_with(app, signal=app.on_worker_init)

    def test_discover(self, *, app):
        app.conf.autodiscover = ['a', 'b', 'c']
        app.conf.origin = 'faust'
        fixup1 = Mock(name='fixup1', autospec=Fixup)
        fixup1.autodiscover_modules.return_value = ['d', 'e']
        app.fixups = [fixup1]
        with patch('faust.app.base.venusian'):
            with patch('importlib.import_module') as import_module:
                app.discover()

                import_module.assert_has_calls(
                    [
                        call('a'),
                        call('b'),
                        call('c'),
                        call('d'),
                        call('e'),
                        call('faust'),
                    ],
                    any_order=True,
                )

    def test_discover__disabled(self, *, app):
        app.conf.autodiscover = False
        app.discover()

    def test_discover__unknown_module(self, *, app):
        app.conf.autodiscover = ['xcxz']
        app.conf.origin = 'faust'
        with patch('faust.app.base.venusian'):
            with pytest.raises(ModuleNotFoundError):
                app.discover()

    def test_discovery_modules__bool(self, *, app):
        app.conf.origin = 'faust'
        app.conf.autodiscover = True
        assert app._discovery_modules() == ['faust']

    def test_discovery_modules__callable(self, *, app):
        app.conf.origin = 'faust'
        app.conf.autodiscover = lambda: ['a', 'b', 'c']
        assert app._discovery_modules() == ['a', 'b', 'c', 'faust']

    def test_discovery_modules__list(self, *, app):
        app.conf.origin = 'faust'
        app.conf.autodiscover = ['a', 'b', 'c']
        assert app._discovery_modules() == ['a', 'b', 'c', 'faust']

    def test_discovery_modules__list_no_origin(self, *, app):
        app.conf.origin = None
        app.conf.autodiscover = ['a', 'b', 'c']
        assert app._discovery_modules() == ['a', 'b', 'c']

    def test_discovery_modules__disabled(self, *, app):
        app.conf.origin = 'faust'
        app.conf.autodiscover = False
        assert app._discovery_modules() == []

    def test_discovery_modules__without_origin(self, *, app):
        app.conf.autodiscover = True
        app.conf.origin = None
        with pytest.raises(ImproperlyConfigured):
            app._discovery_modules()

    def test_discover_ignore(self, *, app):
        with patch('faust.app.base.venusian') as venusian:
            app.conf.origin = 'faust'
            app.conf.autodiscover = ['re', 'collections']
            app.discover(categories=['faust.agent'], ignore=['re', 'faust'])

            assert venusian.Scanner().scan.assert_has_calls([
                call(
                    re,
                    categories=('faust.agent', ),
                    ignore=['re', 'faust'],
                    onerror=app._on_autodiscovery_error,
                ),
                call(
                    faust,
                    categories=('faust.agent', ),
                    ignore=['re', 'faust'],
                    onerror=app._on_autodiscovery_error,
                ),
                call(
                    collections,
                    categories=('faust.agent', ),
                    ignore=['re', 'faust'],
                    onerror=app._on_autodiscovery_error,
                ),
            ], any_order=True) is None

    def test__on_autodiscovery_error(self, *, app):
        with patch('faust.app.base.logger') as logger:
            try:
                raise KeyError('foo')
            except Exception as exc:
                app._on_autodiscovery_error('name')
                logger.warning.assert_called_once_with(
                    ANY, 'name', exc, exc_info=1)

    def test_main(self, *, app):
        with patch('faust.cli.faust.cli') as cli:
            app.finalize = Mock(name='app.finalize')
            app.worker_init = Mock(name='app.worker_init')
            app.discover = Mock(name='app.discover')

            app.conf.autodiscover = False
            with pytest.raises(SystemExit):
                app.main()

            z = app.conf.autodiscover
            assert z is False

            app.finalize.assert_called_once_with()
            app.worker_init.assert_called_once_with()
            cli.assert_called_once_with(app=app)

            with pytest.warns(AlreadyConfiguredWarning):
                app.conf.autodiscover = True
            with pytest.raises(SystemExit):
                app.main()

    def test_channel(self, *, app):
        channel = app.channel()
        assert isinstance(channel, Channel)
        assert isinstance(channel, ChannelT)
        assert app.channel(key_type='x').key_type == 'x'
        assert app.channel(value_type='x').value_type == 'x'
        assert app.channel(maxsize=303).maxsize == 303

    def test_agent(self, *, app):
        with patch('faust.app.base.venusian') as venusian:
            @app.agent()
            async def foo(stream):
                ...

            assert foo.name
            assert app.agents[foo.name] is foo

            venusian.attach.assert_called_once_with(foo, category=SCAN_AGENT)

    @pytest.mark.asyncio
    async def test_on_agent_error(self, *, app):
        app._consumer = None
        agent = Mock(name='agent', autospec=Agent)
        await app._on_agent_error(agent, KeyError())

    @pytest.mark.asyncio
    async def test_on_agent_error__consumer(self, *, app):
        app._consumer = Mock(name='consumer', autospec=Consumer)
        agent = Mock(name='agent', autospec=Agent)
        exc = KeyError()
        await app._on_agent_error(agent, exc)
        app._consumer.on_task_error.assert_called_with(exc)

    @pytest.mark.asyncio
    async def test_on_agent_error__MemoryError(self, *, app):
        app._consumer = Mock(name='consumer', autospec=Consumer)
        app._consumer.on_task_error.side_effect = MemoryError()
        agent = Mock(name='agent', autospec=Agent)
        exc = KeyError()
        with pytest.raises(MemoryError):
            await app._on_agent_error(agent, exc)
        app._consumer.on_task_error.assert_called_with(exc)

    def test_task(self, *, app):
        with patch('faust.app.base.venusian') as venusian:
            @app.task
            async def foo():
                ...

            venusian.attach.assert_called_once_with(foo, category=SCAN_TASK)
            assert foo in app._app_tasks

    @pytest.mark.asyncio
    async def test_task__on_leader(self, *, app):
        @app.task(on_leader=True)
        async def mytask(app):
            return app

        app.is_leader = Mock(return_value=False)
        assert await mytask() is None

        app.is_leader = Mock(return_value=True)
        assert await mytask() is app

    @pytest.mark.asyncio
    async def test_timer(self, *, app):
        did_execute = Mock(name='did_execute')
        app._producer = Mock(name='producer', flush=AsyncMock())

        @app.timer(0.1)
        async def foo():
            did_execute()
            await app.stop()

        await foo()
        did_execute.assert_called_once_with()

    @pytest.mark.asyncio
    async def test_timer__sleep_stopped(self, *, app):
        did_execute = Mock(name='did_execute')
        app.sleep = AsyncMock()

        def on_sleep(seconds, **kwargs):
            app._stopped.set()

        app.sleep.coro.side_effect = on_sleep

        @app.timer(0.1)
        async def foo():
            did_execute()

        await foo()

    @pytest.mark.asyncio
    async def test_timer__on_leader_not_leader(self, *, app):
        did_execute = Mock(name='did_execute')
        app.is_leader = Mock(return_value=False)
        app.sleep = AsyncMock()

        def on_sleep(seconds, **kwargs):
            if app.sleep.call_count >= 3:
                app._stopped.set()
        # cannot use list side_effect arg as it causes
        # StopIteration to be raised.
        app.sleep.coro.side_effect = on_sleep

        @app.timer(300.0, on_leader=True)
        async def foo():
            did_execute()

        await foo()

        assert app.sleep.call_count == 3
        did_execute.assert_not_called()

    @pytest.mark.asyncio
    async def test_timer__on_leader_is_leader(self, *, app):
        did_execute = Mock(name='did_execute')
        app.is_leader = Mock(return_value=True)
        app.sleep = AsyncMock()

        def on_sleep(seconds, **kwargs):
            if app.sleep.call_count >= 3:
                app._stopped.set()
        # cannot use list side_effect arg as it causes
        # StopIteration to be raised.
        app.sleep.coro.side_effect = on_sleep

        @app.timer(300.0, on_leader=True)
        async def foo(app):
            did_execute()

        await foo()

        assert app.sleep.call_count == 3
        assert did_execute.call_count == 2

    @pytest.mark.asyncio
    async def test_crontab(self, *, app):
        did_execute = Mock(name='did_execute')
        app._producer = Mock(name='producer', flush=AsyncMock())

        with patch('faust.app.base.cron') as cron:
            cron.secs_for_next.return_value = 0.1

            @app.crontab('* * * * *')
            async def foo():
                did_execute()
                await app.stop()

            await foo()
            did_execute.assert_called_once_with()

    @pytest.mark.asyncio
    async def test_crontab__on_leader_not_leader(self, *, app):
        did_execute = Mock(name='did_execute')
        with patch('faust.app.base.cron') as cron:
            cron.secs_for_next.return_value = 0.1

            @app.crontab('* * * * *', on_leader=True)
            async def foo():
                did_execute()

        app.is_leader = Mock(return_value=False)
        app.sleep = AsyncMock()

        def on_sleep(seconds, **kwargs):
            if app.sleep.call_count >= 3:
                app._stopped.set()
        app.sleep.coro.side_effect = on_sleep

        await foo()
        assert app.sleep.call_count == 3
        did_execute.assert_not_called()

    @pytest.mark.asyncio
    async def test_crontab__on_leader_is_leader(self, *, app):
        did_execute = Mock(name='did_execute')
        with patch('faust.app.base.cron') as cron:
            cron.secs_for_next.return_value = 0.1

            @app.crontab('* * * * *', on_leader=True)
            async def foo(app):
                did_execute()

        app.is_leader = Mock(return_value=True)
        app.sleep = AsyncMock()

        def on_sleep(seconds, **kwargs):
            if app.sleep.call_count >= 3:
                app._stopped.set()
        app.sleep.coro.side_effect = on_sleep

        await foo()
        assert app.sleep.call_count == 3
        assert did_execute.call_count == 2

    def test_service(self, *, app):

        @app.service
        class Foo(Service):
            ...

        assert Foo in app._extra_services

    def test_is_leader(self, *, app):
        app._leader_assignor = Mock(
            name='_leader_assignor',
            autospec=LeaderAssignor,
        )
        app._leader_assignor.is_leader.return_value = True
        assert app.is_leader()

    def test_Table(self, *, app):
        table = app.Table('name')
        assert app.tables.data['name'] is table

    def test_GlobalTable(self, *, app):
        table = app.GlobalTable('name')
        assert app.tables.data['name'] is table
        assert isinstance(app.tables.data['name'], GlobalTableT)

    def test_SetTable(self, *, app):
        table = app.SetTable('name')
        assert app.tables.data['name'] is table

    def test_SetGlobalTable(self, *, app):
        table = app.SetGlobalTable('name')
        assert app.tables.data['name'] is table
        assert isinstance(app.tables.data['name'], GlobalTableT)

    def test_page(self, *, app):

        with patch('faust.app.base.venusian') as venusian:

            @app.page('/foo')
            async def view(self, request):
                ...

            assert '/foo' in app.web.views

            venusian.attach.assert_called_once_with(view, category=SCAN_PAGE)

    def test_page__with_cors_options(self, *, app):

        with patch('faust.app.base.venusian') as venusian:

            @app.page(
                '/foo',
                cors_options={
                    'http://foo.example.com': ResourceOptions(
                        allow_credentials=True,
                        expose_headers='*',
                        allow_headers='*',
                        max_age=None,
                        allow_methods='*',
                    ),
                },
            )
            async def view(self, request):
                ...

            assert '/foo' in app.web.views

            venusian.attach.assert_called_once_with(view, category=SCAN_PAGE)

    def test_page__view_class_but_not_view(self, *, app):
        with pytest.raises(TypeError):
            @app.page('/foo')
            class Foo:
                ...

    @pytest.mark.asyncio
    async def test_table_route__query_param(self, *, app):
        table = app.Table('foo')
        view = Mock()
        request = Mock()
        app.router.route_req = AsyncMock()

        @app.table_route(table, query_param='q')
        async def routed(self, request):
            return 42

        request.query = {'q': 'KEY'}

        ret = await routed(view, request)
        assert ret is app.router.route_req.coro.return_value

        app.router.route_req.coro.side_effect = SameNode()
        ret = await routed(view, request)
        assert ret == 42

    @pytest.mark.asyncio
    async def test_table_route__match_info(self, *, app):
        table = app.Table('foo')
        view = Mock()
        request = Mock()
        app.router.route_req = AsyncMock()

        @app.table_route(table, match_info='q')
        async def routed(self, request):
            return 42

        request.match_info = {'q': 'KEY'}

        ret = await routed(view, request)
        assert ret is app.router.route_req.coro.return_value

        app.router.route_req.coro.side_effect = SameNode()
        ret = await routed(view, request)
        assert ret == 42

    @pytest.mark.asyncio
    async def test_table_route__exact_key(self, *, app):
        table = app.Table('foo')
        view = Mock()
        request = Mock()
        app.router.route_req = AsyncMock()

        @app.table_route(table, exact_key='active')
        async def routed(self, request):
            return 42

        ret = await routed(view, request)
        assert ret is app.router.route_req.coro.return_value

        app.router.route_req.coro.side_effect = SameNode()
        ret = await routed(view, request)
        assert ret == 42

    def test_table_route__compat_shard_param(self, *, app):
        table = app.Table('foo')
        with pytest.warns(DeprecationWarning):
            @app.table_route(table, shard_param='x')
            async def view(self, request):
                ...

    def test_table_route__query_param_and_shard_param(self, *, app):
        table = app.Table('foo')
        with pytest.warns(DeprecationWarning):
            with pytest.raises(TypeError):
                @app.table_route(table, query_param='q', shard_param='x')
                async def view(self, request):
                    ...

    def test_table_route__missing_param(self, *, app):
        table = app.Table('foo')
        with pytest.raises(TypeError):
            @app.table_route(table)
            async def view(self, request):
                ...

    def test_command(self, *, app):
        @app.command()
        async def foo():
            ...

    def test_command__with_base(self, *, app):
        class MyBase(AppCommand):
            ...

        @app.command(base=MyBase)
        async def foo():
            ...

        assert issubclass(foo, MyBase)

    @pytest.mark.asyncio
    async def test_start_client(self, *, app):
        app.topics.wait_for_subscriptions = AsyncMock()
        app.maybe_start = AsyncMock(name='app.maybe_start')
        await app.start_client()
        assert app.client_only
        app.maybe_start.assert_called_once_with()

    @pytest.mark.asyncio
    async def test_maybe_start_client(self, *, app):
        app.start_client = AsyncMock(name='start_client')
        app._started.set()
        await app.maybe_start_client()
        app.start_client.assert_not_called()

        app._started.clear()
        await app.maybe_start_client()
        app.start_client.assert_called_once_with()

    @pytest.mark.asyncio
    async def test_commit(self, *, app):
        app.topics = Mock(
            name='topics',
            autospec=Conductor,
            commit=AsyncMock(),
        )
        await app.commit({1})
        app.topics.commit.assert_called_with({1})

    def test_Worker(self, *, app):
        app.conf = Mock(name='conf', autospec=Settings)
        worker = app.Worker(loglevel=10)
        app.conf.Worker.assert_called_once_with(app, loglevel=10)
        assert worker is app.conf.Worker()

    def test_create_directories(self, *, app):
        app.conf = Mock(name='conf', autospec=Settings)

        app._create_directories()

        app.conf.datadir.mkdir.assert_called_once_with(exist_ok=True)
        app.conf.appdir.mkdir.assert_called_once_with(exist_ok=True)
        app.conf.tabledir.mkdir.assert_called_once_with(exist_ok=True)

    @pytest.mark.asyncio
    async def test_maybe_start_producer(self, *, app):
        app.in_transaction = True
        assert await app.maybe_start_producer() is app.consumer.transactions

        app.in_transaction = False
        app.producer = Mock(maybe_start=AsyncMock())
        assert await app.maybe_start_producer() is app.producer
        app.producer.maybe_start.coro.assert_called_once_with()

    def test_repr(self, *, app):
        assert repr(app)

    def test_repr__unfinialized(self, *, app):
        app._conf = None
        assert repr(app)

    def test_monitor(self, *, app):
        assert app._monitor is None
        app.conf.Monitor = Mock(
            name='Monitor',
            return_value=Mock(autospec=Monitor),
        )
        monitor = app.monitor
        app.conf.Monitor.assert_called_once_with(
            loop=app.loop, beacon=app.beacon)
        assert monitor is app.conf.Monitor()
        assert app.monitor is monitor
        assert app._monitor is monitor

        monitor2 = app.monitor = Mock(
            name='monitor2',
            autospec=Monitor,
        )
        assert app._monitor is monitor2
        assert app.monitor is monitor2

    def test_fetcher(self, *, app):
        app.transport.Fetcher = Mock(
            name='Fetcher',
            return_value=Mock(autospec=Fetcher),
        )
        fetcher = app._fetcher
        app.transport.Fetcher.assert_called_once_with(
            app, loop=app.loop, beacon=app.consumer.beacon,
        )
        assert fetcher is app.transport.Fetcher()

    def test_reply_consumer(self, *, app):
        with patch('faust.app.base.ReplyConsumer') as ReplyConsumer:
            reply_consumer = app._reply_consumer
            ReplyConsumer.assert_called_once_with(
                app, loop=app.loop, beacon=app.beacon)
            assert reply_consumer is ReplyConsumer()

    def test_label(self, *, app):
        assert app.label

    def test_cache(self, *, app):
        assert app.cache
        assert app.cache is app._cache
        obj = object()
        app.cache = obj
        assert app.cache is obj

    def test_http_client(self, *, app):
        app.conf.HttpClient = Mock(name='HttpClient')
        assert app._http_client is None
        client = app.http_client
        app.conf.HttpClient.assert_called_once_with()
        assert client is app.conf.HttpClient()
        assert app._http_client is client
        assert app.http_client is client

        obj = object()
        app.http_client = obj
        assert app.http_client is obj

    def test_leader_assignor(self, *, app):
        leader_assignor = app._leader_assignor
        assert isinstance(leader_assignor, LeaderAssignorT)
        assert isinstance(leader_assignor, LeaderAssignor)
        assert leader_assignor.app is app
        assert leader_assignor.beacon.parent is app.beacon


class test_AppConfiguration:

    def test_conf__before_finalized(self, *, monkeypatch, app):
        app.finalized = False
        monkeypatch.setattr('faust.app.base.STRICT', False)
        app.conf.id
        monkeypatch.setattr('faust.app.base.STRICT', True)
        with pytest.raises(ImproperlyConfigured):
            app.conf.id

    def test_set_conf(self, *, app):
        conf = Mock(name='conf', autospec=Settings)
        app.conf = conf
        assert app.conf is conf

    @pytest.mark.parametrize('config_source', [
        ConfigClass,
        CONFIG_DICT,
        CONFIG_PATH,
    ])
    def test_config_From_object(self, config_source, *, app):
        on_before = app.on_before_configured.connect(Mock(name='on_before'))
        on_config = app.on_configured.connect(Mock(name='on_config'))
        on_after = app.on_after_configured.connect(Mock(name='on_after'))

        app.configured = False
        app.finalized = False
        app.config_from_object(config_source)
        app._conf = None
        app.finalized = True

        app.configured = False
        with pytest.warns(AlreadyConfiguredWarning):
            app.config_from_object(config_source)

        assert app._config_source is config_source
        app._config_source = None

        app.configured = True
        with pytest.warns(AlreadyConfiguredWarning):
            app.config_from_object(config_source)

        assert app._config_source is config_source
        on_before.assert_called_with(app, signal=app.on_before_configured)
        on_config.assert_called_with(app, app.conf, signal=app.on_configured)
        on_after.assert_called_with(app, signal=app.on_after_configured)
        assert app.conf.broker == [URL('kafka://foo')]
        assert app.conf.stream_buffer_maxsize == 1

    def test_finalize__no_id(self, *, app):
        with pytest.warns(AlreadyConfiguredWarning):
            app.conf.id = None
        app.finalized = False
        with pytest.raises(ImproperlyConfigured):
            app.finalize()

    def test_load_settings_from_source__no_attribute(self, *, app):
        with pytest.raises(AttributeError):
            app._load_settings_from_source('faust.app.base.NOTEXIST')
        assert app._load_settings_from_source(
            'faust.app.base.NOTEXIST', silent=True) == {}

    def test_load_settings_from_source__import_error(self, *, app):
        with pytest.raises(ImportError):
            app._load_settings_from_source('zxzc')
        assert app._load_settings_from_source('zxzc', silent=True) == {}

    def test_load_settings_with_compat_and_new_settings(self, *, app):
        config = {
            'client_id': 'foo',
            'broker_client_id': 'bar',
        }
        with pytest.warns(AlreadyConfiguredWarning):
            with pytest.raises(ImproperlyConfigured):
                app.config_from_object(config)

    @pytest.mark.asyncio
    async def test_wait_for_table_recovery__producer_only(self, *, app):
        app.producer_only = True
        assert await app._wait_for_table_recovery_completed() is False

    @pytest.mark.asyncio
    async def test_wait_for_table_recovery__client_only(self, *, app):
        app.client_only = True
        assert await app._wait_for_table_recovery_completed() is False



"""Attachments - Deprecated module.

Attachments were used before transactions support.
"""
import asyncio
import typing

from collections import defaultdict
from heapq import heappop, heappush
from typing import (
    Awaitable, Callable, Iterator, List,
    MutableMapping, NamedTuple, Union, cast,
)

from mode.utils.objects import Unordered, cached_property

from faust.streams import current_event
from faust.types import AppT, ChannelT, CodecArg, RecordMetadata, SchemaT, TP
from faust.types.core import HeadersArg, K, V
from faust.types.tuples import FutureMessage, Message, MessageSentCallback

if typing.TYPE_CHECKING:
    from faust.events import Event as _Event
else:
    class _Event: ...       # noqa

__all__ = ['Attachment', 'Attachments']


class Attachment(NamedTuple):
    """Message attached to offset in source topic.

    The message will be published once that offset in the source
    topic is committed.
    """

    # Tuple used in heapq entry for Attachments._pending
    # These are used to delay producing of messages until source offset is
    # committed:
    #
    # @app.agent(source_topic)
    # async def process(stream):
    #    async for value in stream:
    #        await other_topic.send(value)  # does not send here!
    #
    # sending of the message is attached to the offset of value in the source
    # topic, so that when the event is acked, only then do we send the
    # message.  This gives better consistency: if we reprocess the event
    # we don't send messages twice.
    #
    # Note though: we need Kafka transactions to cover all cases of
    # inconsistencies.

    offset: int
    message: Unordered[FutureMessage]


class Attachments:
    """Attachments manager."""

    app: AppT

    # Mapping used to attach messages to a source message such that
    # only when the source message is acked, only then do we publish
    # its attached messages.
    #
    # The mapping maintains one list for each TopicPartition,
    # where the lists are used as heap queues containing tuples
    # of ``(source_message_offset, FutureMessage)``.
    _pending: MutableMapping[TP, List[Attachment]]

    def __init__(self, app: AppT) -> None:
        self.app = app
        self._pending = defaultdict(list)

    @cached_property
    def enabled(self) -> bool:
        """Return :const:`True` if attachments are enabled."""
        return self.app.conf.stream_publish_on_commit

    async def maybe_put(self,
                        channel: Union[ChannelT, str],
                        key: K = None,
                        value: V = None,
                        partition: int = None,
                        timestamp: float = None,
                        headers: HeadersArg = None,
                        schema: SchemaT = None,
                        key_serializer: CodecArg = None,
                        value_serializer: CodecArg = None,
                        callback: MessageSentCallback = None,
                        force: bool = False) -> Awaitable[RecordMetadata]:
        """Attach message to source topic offset.

        This will send the message immediately if attachments
        are disabled.
        """
        # XXX The concept of attaching should be deprecated when we
        # have Kafka transaction support (:kip:`KIP-98`).
        # This is why the interface related to attaching is private.

        # attach message to current event if there is one.
        send: Callable = self.app.send
        if self.enabled and not force:
            event = current_event()
            if event is not None:
                return cast(_Event, event)._attach(
                    channel,
                    key,
                    value,
                    partition=partition,
                    timestamp=timestamp,
                    headers=headers,
                    schema=schema,
                    key_serializer=key_serializer,
                    value_serializer=value_serializer,
                    callback=callback,
                )
        return await send(
            channel,
            key,
            value,
            partition=partition,
            timestamp=timestamp,
            headers=headers,
            schema=schema,
            key_serializer=key_serializer,
            value_serializer=value_serializer,
            callback=callback,
        )

    def put(self,
            message: Message,
            channel: Union[str, ChannelT],
            key: K,
            value: V,
            partition: int = None,
            timestamp: float = None,
            headers: HeadersArg = None,
            schema: SchemaT = None,
            key_serializer: CodecArg = None,
            value_serializer: CodecArg = None,
            callback: MessageSentCallback = None) -> Awaitable[RecordMetadata]:
        """Attach message to source topic offset."""
        # This attaches message to be published when source message' is
        # acknowledged.  To be replaced by transactions in :kip:`KIP-98`.

        # get heap queue for this TopicPartition
        # items in this list are ``(source_offset, Unordered[FutureMessage])``
        # tuples.
        buf = self._pending[message.tp]
        chan = self.app.topic(channel) if isinstance(channel, str) else channel
        fut = chan.as_future_message(
            key, value, partition, timestamp, headers,
            schema, key_serializer, value_serializer, callback)
        # Note: Since FutureMessage have members that are unhashable
        # we wrap it in an Unordered object to stop heappush from crashing.
        # Unordered simply orders by random order, which is fine
        # since offsets are always unique.
        heappush(buf, Attachment(message.offset, Unordered(fut)))
        return fut

    async def commit(self, tp: TP, offset: int) -> None:
        """Publish all messaged attached to topic partition and offset."""
        await asyncio.wait(
            await self.publish_for_tp_offset(tp, offset),
            return_when=asyncio.ALL_COMPLETED,
            loop=self.app.loop,
        )

    async def publish_for_tp_offset(
            self, tp: TP, offset: int) -> List[Awaitable[RecordMetadata]]:
        """Publish messages attached to topic partition and offset."""
        # publish pending messages attached to this TP+offset

        # make shallow copy to allow concurrent modifications (append)
        attached = list(self._attachments_for(tp, offset))
        return [
            await fut.message.channel.publish_message(fut, wait=False)
            for fut in attached
        ]

    def _attachments_for(self, tp: TP,
                         commit_offset: int) -> Iterator[FutureMessage]:
        # Return attached messages for TopicPartition within committed offset.
        attached = self._pending.get(tp)
        while attached:
            # get the entry with the smallest offset in this TP
            entry = heappop(attached)

            # if the entry offset is smaller or equal to the offset
            # being committed
            if entry[0] <= commit_offset:
                # we use it by extracting the FutureMessage
                # from Attachment tuple, where entry.message is
                # Unordered[FutureMessage].
                yield entry.message.value
            else:
                # else we put it back and exit (this was the smallest offset).
                heappush(attached, entry)
                break



from faust.windows import TumblingWindow


class test_TumblingWindow:

    def test_tumbling_window_has_just_one_range(self):
        tumbling = TumblingWindow(10)
        assert len(tumbling.ranges(0)) == 1
        assert len(tumbling.ranges(5)) == 1
        assert len(tumbling.ranges(10)) == 1

    def test_end_range_in_tumbling_window_is_within_range(self):
        tumbling = TumblingWindow(10)

        # tumbling windows only have one range
        base_range = tumbling.ranges(0)[0]
        base_range_end = base_range[1]

        compare_range = tumbling.ranges(base_range_end)[0]

        assert base_range[0] == compare_range[0]
        assert base_range[1] == compare_range[1]

    def test_earliest_and_current_range_are_the_same(self):
        size = 57
        timestamp = 456
        window = TumblingWindow(size)

        assert window.current(timestamp) == window.earliest(timestamp)



"""JSON utilities."""
import datetime
import enum
import typing
import uuid

from collections import Counter, deque
from decimal import Decimal
from typing import (
    Any,
    Callable,
    Iterable,
    List,
    Mapping,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)

__all__ = [
    'JSONEncoder',
    'dumps',
    'loads',
    'str_to_decimal',
]

if typing.TYPE_CHECKING:
    import orjson
else:  # pragma: no cover
    orjson = None  # noqa

DEFAULT_TEXTUAL_TYPES: List[Type] = [Decimal, uuid.UUID, bytes]

T = TypeVar('T')

TypeTuple = Tuple[Type[T], ...]
IsInstanceArg = Union[Type[Any], TypeTuple[Any]]

try:  # pragma: no cover
    from django.utils.functional import Promise
    DJANGO_TEXTUAL_TYPES = [Promise]
except ImportError:  # pragma: no cover
    DJANGO_TEXTUAL_TYPES = []

TEXTUAL_TYPES: TypeTuple[Any] = tuple(
    DEFAULT_TEXTUAL_TYPES + DJANGO_TEXTUAL_TYPES)

try:  # pragma: no cover
    import simplejson as json

    # simplejson converts Decimal to float by default, i.e. before
    # we have a chance to override it using Encoder.default.
    _JSON_DEFAULT_KWARGS = {
        'use_decimal': False,
        'namedtuple_as_object': False,
    }

except ImportError:  # pragma: no cover
    import json  # type: ignore
    _JSON_DEFAULT_KWARGS = {}

#: Max length for string to be converted to decimal.
DECIMAL_MAXLEN = 1000

#: Types that we convert to lists.
SEQUENCE_TYPES: TypeTuple[Iterable] = (set, deque)

DateTypeTuple = Tuple[Union[Type[datetime.date], Type[datetime.time]], ...]
DatetimeTypeTuple = Tuple[
    Union[
        Type[datetime.time],
        Type[datetime.datetime],
    ],
    ...,
]

#: Types that we convert o mapping.
MAPPING_TYPES: TypeTuple[Mapping] = (Counter,)

#: Types that are datetimes and dates (-> .isoformat())
DATE_TYPES: DateTypeTuple = (datetime.date, datetime.time)

#: Types we use `return obj.value` for (Enum)
VALUE_DELEGATE_TYPES: TypeTuple[enum.Enum] = (enum.Enum,)

HAS_TIME: DatetimeTypeTuple = (datetime.datetime, datetime.time)


def str_to_decimal(s: str, maxlen: int = DECIMAL_MAXLEN) -> Optional[Decimal]:
    """Convert string to :class:`~decimal.Decimal`.

    Args:
        s (str): Number to convert.
        maxlen (int): Max length of string.  Default is 100.

    Raises:
        ValueError: if length exceeds maximum length, or if value is not
            a valid number (e.g. Inf, NaN or sNaN).

    Returns:
        Decimal: Converted number.
    """
    if s is None:
        return None
    if len(s) > maxlen:
        raise ValueError(
            f'string of length {len(s)} is longer than limit ({maxlen})')
    v = Decimal(s)
    if not v.is_finite():  # check for Inf/NaN/sNaN/qNaN
        raise ValueError(f'Illegal value in decimal: {s!r}')
    return v


def on_default(o: Any,
               *,
               sequences: TypeTuple[Iterable] = SEQUENCE_TYPES,
               maps: TypeTuple[Mapping] = MAPPING_TYPES,
               dates: DateTypeTuple = DATE_TYPES,
               value_delegate: TypeTuple[enum.Enum] = VALUE_DELEGATE_TYPES,
               has_time: DatetimeTypeTuple = HAS_TIME,
               _isinstance: Callable[[Any, IsInstanceArg], bool] = isinstance,
               _dict: Callable = dict,
               _str: Callable[[Any], str] = str,
               _list: Callable = list,
               textual: TypeTuple[Any] = TEXTUAL_TYPES) -> Any:
    if _isinstance(o, textual):
        return _str(o)
    elif _isinstance(o, maps):
        return _dict(o)
    elif _isinstance(o, dates):
        if not _isinstance(o, has_time):
            o = datetime.datetime(o.year, o.month, o.day, 0, 0, 0, 0)
        r = o.isoformat()
        if r.endswith('+00:00'):
            r = r[:-6] + 'Z'
        return r
    elif isinstance(o, value_delegate):
        return o.value
    elif isinstance(o, sequences):
        return _list(o)
    else:
        to_json = getattr(o, '__json__', None)
        if to_json is not None:
            return to_json()
        raise TypeError(
            f'JSON cannot serialize {type(o).__name__!r}: {o!r}')


class JSONEncoder(json.JSONEncoder):
    """Faust customized :class:`json.JSONEncoder`.

    Our version supports additional types like :class:`~uuid.UUID`, and
    importantly includes microsecond information in datetimes.
    """

    def default(self, o: Any, *,
                callback: Callable[[Any], Any] = on_default) -> Any:
        """Try to convert non-built-in json type to json."""
        return callback(o)


if orjson is not None:  # pragma: no cover

    def dumps(obj: Any,
              json_dumps: Callable = orjson.dumps,
              cls: Type[JSONEncoder] = JSONEncoder,
              **kwargs: Any) -> str:
        """Serialize to json."""
        return json_dumps(
            obj,
            default=on_default,
            option=orjson.OPT_NAIVE_UTC,
        )

    def loads(s: str,
              json_loads: Callable = orjson.loads,
              **kwargs: Any) -> Any:
        """Deserialize json string."""
        return json_loads(s)
else:

    def dumps(obj: Any,
              json_dumps: Callable = json.dumps,
              cls: Type[JSONEncoder] = JSONEncoder,
              **kwargs: Any) -> str:
        """Serialize to json.  See :func:`json.dumps`."""
        return json_dumps(obj, cls=cls, **dict(_JSON_DEFAULT_KWARGS, **kwargs))

    def loads(s: str,
              json_loads: Callable = json.loads,
              **kwargs: Any) -> Any:
        """Deserialize json string.  See :func:`json.loads`."""
        return json_loads(s, **kwargs)



"""Transport registry."""
from typing import Type
from mode.utils.imports import FactoryMapping
from faust.types import TransportT

__all__ = ['by_name', 'by_url']

TRANSPORTS: FactoryMapping[Type[TransportT]] = FactoryMapping(
    aiokafka='faust.transport.drivers.aiokafka:Transport',
    confluent='faust.transport.drivers.confluent:Transport',
    kafka='faust.transport.drivers.aiokafka:Transport',
)
TRANSPORTS.include_setuptools_namespace('faust.transports')
by_name = TRANSPORTS.by_name
by_url = TRANSPORTS.by_url



from faust.windows import SlidingWindow


class test_SlidingWindow:

    def test_constructor(self):
        x = SlidingWindow(10.1, 20.2, 30.3)
        assert x.before == 10.1
        assert x.after == 20.2
        assert x.expires == 30.3

    def test_has_ranges_including_the_value(self):
        size = 10
        step = 5
        timestamp = 6

        window = SlidingWindow(size, step, 30)

        window_ranges = window.ranges(timestamp)
        assert len(window_ranges) == 1
        for range in window_ranges:
            assert range[0] <= timestamp
            assert range[1] > timestamp

    def test_current_range_is_latest_range(self):
        size = 57
        step = 23
        timestamp = 456

        window = SlidingWindow(size, step, 30)
        ranges = window.ranges(timestamp)
        current_range = window.current(timestamp)

        assert current_range == ranges[-1]

    def test_earliest_range_is_first_range(self):
        size = 100
        step = 15
        timestamp = 3223

        window = SlidingWindow(size, step, 30)
        ranges = window.ranges(timestamp)
        earliest_range = window.earliest(timestamp)

        assert earliest_range == ranges[0]

    def test_non_stale_timestamp(self):
        size = 10
        step = 5
        expires = 20

        now_timestamp = 60

        window = SlidingWindow(size, step, expires)
        for time in range(now_timestamp - expires + 1, now_timestamp):
            assert window.stale(time, now_timestamp) is False

    def test_delta(self):
        size = 10
        step = 5
        expires = 20

        now_timestamp = 60
        window = SlidingWindow(size, step, expires)
        assert window.delta(now_timestamp, 30) == (20, 35)

    def test_stale_timestamp(self):
        size = 10
        step = 5
        expires = 20
        now_timestamp = 60

        window = SlidingWindow(size, step, expires)
        for time in range(0, now_timestamp - expires):
            print(f'TIME: {time} NOW TIMESTAMP: {now_timestamp}')
            assert window.stale(time, now_timestamp) is True



from datetime import datetime, timedelta, timezone
import pytest
from mode.utils.mocks import ContextMock, patch
from faust.livecheck.models import TestExecution
from faust.livecheck.runners import TestRunner


@pytest.fixture()
def livecheck(*, app):
    return app.LiveCheck()


@pytest.fixture()
def execution():
    now = datetime.now().astimezone(timezone.utc)
    expires = now + timedelta(hours=3)
    return TestExecution(
        id='id',
        case_name='t.examples.test_foo',
        timestamp=now,
        test_args=('foo',),
        test_kwargs={'kw1': 1.03},
        expires=expires,
    )


@pytest.fixture()
def case(*, livecheck):
    @livecheck.case()
    class test_foo(livecheck.Case):

        async def run(self, arg1, kw1=None):
            assert arg1 == 'foo'
            assert kw1 == 1.03
            assert True
    return test_foo


@pytest.fixture()
def runner(*, execution, case):
    return TestRunner(case, execution, started=100.0)


@pytest.yield_fixture()
def current_test_stack():
    with patch('faust.livecheck.case.current_test_stack') as cts:
        cts.push = ContextMock()
        yield cts


@pytest.yield_fixture()
def current_execution_stack():
    with patch('faust.livecheck.case.current_execution_stack') as ces:
        ces.push = ContextMock()
        yield ces



import inspect
from typing import Any, AnyStr, TypeVar, get_type_hints
from sphinx.util.inspect import Signature

try:
    from typing import GenericMeta  # Py3.7
except ImportError:
    from typing import Generic as GenericMeta  # Py3.7


def format_annotation(annotation):
    if inspect.isclass(annotation) and annotation.__module__ == 'builtins':
        if annotation.__qualname__ == 'NoneType':
            return '``None``'
        else:
            return ':py:class:`{}`'.format(annotation.__qualname__)

    annotation_cls = (
        annotation if inspect.isclass(annotation) else type(annotation))
    if annotation_cls.__module__ == 'typing':
        params = None
        prefix = ':py:class:'
        extra = ''
        class_name = annotation_cls.__qualname__
        if annotation is Any:
            return ':py:data:`~typing.Any`'
        elif annotation is AnyStr:
            return ':py:data:`~typing.AnyStr`'
        elif isinstance(annotation, TypeVar):
            return f'\\{annotation!r}'
        elif class_name in ('Union', '_Union'):
            prefix = ':py:data:'
            class_name = 'Union'
            if hasattr(annotation, '__union_params__'):
                params = annotation.__union_params__
            else:
                params = annotation.__args__

            if params and len(params) == 2 and (
                    hasattr(params[1], '__qualname__') and
                    params[1].__qualname__ == 'NoneType'):
                class_name = 'Optional'
                params = (params[0],)
        elif annotation_cls.__qualname__ == 'Tuple' and (
                hasattr(annotation, '__tuple_params__')):
            params = annotation.__tuple_params__
            if annotation.__tuple_use_ellipsis__:
                params += (Ellipsis,)
        elif annotation_cls.__qualname__ == 'Callable':
            prefix = ':py:data:'
            arg_annotations = result_annotation = None
            if hasattr(annotation, '__result__'):
                arg_annotations = annotation.__args__
                result_annotation = annotation.__result__
            elif getattr(annotation, '__args__', None) is not None:
                arg_annotations = annotation.__args__[:-1]
                result_annotation = annotation.__args__[-1]

            if arg_annotations in (Ellipsis, (Ellipsis,)):
                params = [Ellipsis, result_annotation]
            elif arg_annotations is not None:
                params = [
                    '\\[{}]'.format(
                        ', '.join(format_annotation(param)
                                  for param in arg_annotations)),
                    result_annotation,
                ]
        elif hasattr(annotation, 'type_var'):
            # Type alias
            class_name = annotation.name
            params = (annotation.type_var,)
        elif getattr(annotation, '__args__', None) is not None:
            params = annotation.__args__
        elif hasattr(annotation, '__parameters__'):
            params = annotation.__parameters__

        if params:
            extra = '\\[{}]'.format(
                ', '.join(format_annotation(param) for param in params))

        return '{}`~typing.{}`{}'.format(prefix, class_name, extra)
    elif annotation is Ellipsis:
        return '...'
    elif inspect.isclass(annotation):
        extra = ''
        if isinstance(annotation, GenericMeta):
            extra = '\\[{}]'.format(', '.join(
                format_annotation(param)
                for param in annotation.__parameters__))

        return ':py:class:`~{}.{}`{}'.format(
            annotation.__module__,
            annotation.__qualname__,
            extra)
    else:
        return str(annotation)


def process_signature(app, what: str, name: str, obj, options,
                      signature, return_annotation):
    if callable(obj) and getattr(obj, '__annotations__', None):
        if what in ('class', 'exception'):
            obj = getattr(obj, '__init__')

        bound_method = what in ('method', 'class', 'exception')
        obj = inspect.unwrap(obj)
        try:
            sig = Signature(obj, bound_method=bound_method)
        except (TypeError, ValueError):
            return

        formatted_args = sig.format_args()
        return formatted_args, None


def process_docstring(app, what, name, obj, options, lines):
    if isinstance(obj, property):
        obj = obj.fget

    if callable(obj):
        if what in ('class', 'exception'):
            obj = getattr(obj, '__init__')

        obj = inspect.unwrap(obj)
        try:
            type_hints = get_type_hints(obj)
        except (AttributeError, TypeError):
            # Introspecting a slot wrapper will raise TypeError
            return

        for argname, annotation in type_hints.items():
            formatted_annotation = format_annotation(annotation)

            if argname == 'return':
                if what in ('class', 'exception'):
                    # Don't add return type None from __init__()
                    continue

                insert_index = len(lines)
                for i, line in enumerate(lines):
                    if line.startswith(':rtype:'):
                        insert_index = None
                        break
                    elif line.startswith((':return:', ':returns:')):
                        insert_index = i
                        break

                if insert_index is not None:
                    lines.insert(insert_index, ':rtype: {}'.format(
                        formatted_annotation))
            else:
                searchfor = ':param {}:'.format(argname)
                for i, line in enumerate(lines):
                    if line.startswith(searchfor):
                        lines.insert(i, ':type {}: {}'.format(
                            argname, formatted_annotation))
                        break
                searchfor2 = ':keyword {}:'.format(argname)
                for i, line in enumerate(lines):
                    line.replace(':keyword ', ':param ')
                    if line.startswith(searchfor2):
                        lines.insert(i, ':type {}: {}'.format(
                            argname, formatted_annotation))
                        break


def setup(app):
    app.connect('autodoc-process-signature', process_signature)
    app.connect('autodoc-process-docstring', process_docstring)
    return {'parallel_read_safe': True}



# Generated by Django 2.1.4 on 2018-12-24 12:13

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Account',
            fields=[
                ('id', models.AutoField(auto_created=True,
                                        primary_key=True,
                                        serialize=False,
                                        verbose_name='ID')),
                ('name', models.CharField(max_length=100,
                                          verbose_name='name')),
                ('score', models.DecimalField(decimal_places=1000,
                                              default=0.0,
                                              max_digits=1000,
                                              verbose_name='score')),
                ('active', models.BooleanField(default=True,
                                               verbose_name='active')),
            ],
            options={
                'verbose_name': 'account',
                'verbose_name_plural': 'accounts',
            },
        ),
    ]



"""Monitor - sensor tracking metrics."""
import asyncio

from collections import deque
from http import HTTPStatus
from statistics import median
from time import monotonic
from typing import (
    Any,
    Callable,
    Dict,
    Mapping,
    MutableMapping,
    Optional,
    Tuple,
    cast,
)

from mode import Service, label
from mode.utils.objects import KeywordReduce
from mode.utils.typing import Counter, Deque

from faust import web
from faust.types import AppT, CollectionT, EventT, StreamT, TopicT
from faust.types.assignor import PartitionAssignorT
from faust.types.tuples import Message, PendingMessage, RecordMetadata, TP
from faust.types.transports import ConsumerT, ProducerT
from faust.utils.functional import deque_pushpopmax

from .base import Sensor

__all__ = ['TableState', 'Monitor']

MAX_AVG_HISTORY = 100
MAX_COMMIT_LATENCY_HISTORY = 30
MAX_SEND_LATENCY_HISTORY = 30
MAX_ASSIGNMENT_LATENCY_HISTORY = 30

TPOffsetMapping = MutableMapping[TP, int]
PartitionOffsetMapping = MutableMapping[int, int]
TPOffsetDict = MutableMapping[str, PartitionOffsetMapping]


class TableState(KeywordReduce):
    """Represents the current state of a table."""

    #: The table this object records statistics for.
    # the attribute here cannot be None, but need to exist on the
    # class so that Sphinx finds it, so we cast the None to its type.
    table: CollectionT = cast(CollectionT, None)

    #: Number of times a key has been retrieved from this table.
    keys_retrieved: int = 0

    #: Number of times a key has been created/changed in this table.
    keys_updated: int = 0

    #: Number of times a key has been deleted from this table.
    keys_deleted: int = 0

    def __init__(self,
                 table: CollectionT,
                 *,
                 keys_retrieved: int = 0,
                 keys_updated: int = 0,
                 keys_deleted: int = 0) -> None:
        self.table: CollectionT = table
        self.keys_retrieved = keys_retrieved
        self.keys_updated = keys_updated
        self.keys_deleted = keys_deleted

    def asdict(self) -> Mapping:
        """Return table state as dictionary."""
        return {
            'keys_retrieved': self.keys_retrieved,
            'keys_updated': self.keys_updated,
            'keys_deleted': self.keys_deleted,
        }

    def __reduce_keywords__(self) -> Mapping:
        return {**self.asdict(), 'table': self.table}


class Monitor(Sensor, KeywordReduce):
    """Default Faust Sensor.

    This is the default sensor, recording statistics about
    events, etc.
    """

    #: Max number of total run time values to keep to build average.
    max_avg_history: int = MAX_AVG_HISTORY

    #: Max number of commit latency numbers to keep.
    max_commit_latency_history: int = MAX_COMMIT_LATENCY_HISTORY

    #: Max number of send latency numbers to keep.
    max_send_latency_history: int = MAX_SEND_LATENCY_HISTORY

    #: Max number of assignment latency numbers to keep.
    max_assignment_latency_history: int = MAX_ASSIGNMENT_LATENCY_HISTORY

    #: Mapping of tables
    tables: MutableMapping[str, TableState] = cast(
        MutableMapping[str, TableState], None)

    #: Number of messages currently being processed.
    messages_active: int = 0

    #: Number of messages processed in total.
    messages_received_total: int = 0

    #: Count of messages received by topic
    messages_received_by_topic: Counter[str] = cast(Counter[str], None)

    #: Number of messages being processed this second.
    messages_s: int = 0

    #: Number of messages sent in total.
    messages_sent: int = 0

    #: Number of messages sent by topic.
    messages_sent_by_topic: Counter[str] = cast(Counter[str], None)

    #: Number of events currently being processed.
    events_active: int = 0

    #: Number of events processed in total.
    events_total: int = 0

    #: Number of events being processed this second.
    events_s: int = 0

    #: Count of events processed by stream
    events_by_stream: Counter[str] = cast(Counter[str], None)

    #: Count of events processed by task
    events_by_task: Counter[str] = cast(Counter[str], None)

    #: Average event runtime over the last second.
    events_runtime_avg: float = 0.0

    #: Deque of run times used for averages
    events_runtime: Deque[float] = cast(Deque[float], None)

    #: Deque of commit latency values
    commit_latency: Deque[float] = cast(Deque[float], None)

    #: Deque of send latency values
    send_latency: Deque[float] = cast(Deque[float], None)

    #: Deque of assignment latency values.
    assignment_latency: Deque[float] = cast(Deque[float], None)

    #: Counter of times a topics buffer was full
    topic_buffer_full: Counter[TopicT] = cast(Counter[TopicT], None)

    #: Arbitrary counts added by apps
    metric_counts: Counter[str] = cast(Counter[str], None)

    #: Last committed offsets by TopicPartition
    tp_committed_offsets: TPOffsetMapping = cast(TPOffsetMapping, None)

    #: Last read offsets by TopicPartition
    tp_read_offsets: TPOffsetMapping = cast(TPOffsetMapping, None)

    #: Log end offsets by TopicPartition
    tp_end_offsets: TPOffsetMapping = cast(TPOffsetMapping, None)

    #: Number of produce operations that ended in error.
    send_errors = 0

    #: Number of partition assignments completed.
    assignments_completed = 0

    #: Number of partitions assignments that failed.
    assignments_failed = 0

    #: Number of rebalances seen by this worker.
    rebalances = 0

    #: Deque of previous n rebalance return latencies.
    rebalance_return_latency: Deque[float] = cast(Deque[float], None)

    #: Deque of previous n rebalance end latencies.
    rebalance_end_latency: Deque[float] = cast(Deque[float], None)

    #: Average rebalance return latency.
    rebalance_return_avg: float = .0

    #: Average rebalance end latency.
    rebalance_end_avg: float = .0

    #: Counter of returned HTTP status codes.
    http_response_codes: Counter[HTTPStatus] = cast(Counter[HTTPStatus], None)

    #: Deque of previous n HTTP request->response latencies.
    http_response_latency: Deque[float] = cast(Deque[float], None)

    #: Average request->response latency.
    http_response_latency_avg: float = .0

    stream_inbound_time: Dict[TP, float] = cast(Dict[TP, float], None)

    def __init__(self,
                 *,
                 max_avg_history: int = None,
                 max_commit_latency_history: int = None,
                 max_send_latency_history: int = None,
                 max_assignment_latency_history: int = None,
                 messages_sent: int = 0,
                 tables: MutableMapping[str, TableState] = None,
                 messages_active: int = 0,
                 events_active: int = 0,
                 messages_received_total: int = 0,
                 messages_received_by_topic: Counter[str] = None,
                 events_total: int = 0,
                 events_by_stream: Counter[StreamT] = None,
                 events_by_task: Counter[asyncio.Task] = None,
                 events_runtime: Deque[float] = None,
                 commit_latency: Deque[float] = None,
                 send_latency: Deque[float] = None,
                 assignment_latency: Deque[float] = None,
                 events_s: int = 0,
                 messages_s: int = 0,
                 events_runtime_avg: float = 0.0,
                 topic_buffer_full: Counter[TopicT] = None,
                 rebalances: int = None,
                 rebalance_return_latency: Deque[float] = None,
                 rebalance_end_latency: Deque[float] = None,
                 rebalance_return_avg: float = 0.0,
                 rebalance_end_avg: float = 0.0,
                 time: Callable[[], float] = monotonic,
                 http_response_codes: Counter[HTTPStatus] = None,
                 http_response_latency: Deque[float] = None,
                 http_response_latency_avg: float = 0.0,
                 **kwargs: Any) -> None:
        if max_avg_history is not None:
            self.max_avg_history = max_avg_history
        if max_commit_latency_history is not None:
            self.max_commit_latency_history = max_commit_latency_history
        if max_send_latency_history is not None:
            self.max_send_latency_history = max_send_latency_history
        if max_assignment_latency_history is not None:
            self.max_assignment_latency_history = (
                max_assignment_latency_history)
        if rebalances is not None:
            self.rebalances = rebalances

        self.tables = {} if tables is None else tables
        self.commit_latency = (
            deque() if commit_latency is None else commit_latency)
        self.send_latency = deque() if send_latency is None else send_latency
        self.assignment_latency = (
            deque() if assignment_latency is None else assignment_latency)
        self.rebalance_return_latency = (
            deque() if rebalance_return_latency is None
            else rebalance_return_latency)
        self.rebalance_end_latency = (
            deque() if rebalance_end_latency is None
            else rebalance_end_latency)
        self.rebalance_return_avg = rebalance_return_avg
        self.rebalance_end_avg = rebalance_end_avg

        self.messages_active = messages_active
        self.messages_received_total = messages_received_total
        self.messages_received_by_topic = Counter()
        self.messages_sent = messages_sent
        self.messages_sent_by_topic = Counter()
        self.messages_s = messages_s

        self.events_active = events_active
        self.events_total = events_total
        self.events_by_task = Counter()
        self.events_by_stream = Counter()
        self.events_s = events_s
        self.events_runtime_avg = events_runtime_avg
        self.events_runtime = (
            deque() if events_runtime is None else events_runtime)
        self.topic_buffer_full = Counter()
        self.time: Callable[[], float] = time

        self.http_response_codes = Counter()
        self.http_response_latency = deque()
        self.http_response_latency_avg = http_response_latency_avg

        self.metric_counts = Counter()

        self.tp_committed_offsets = {}
        self.tp_read_offsets = {}
        self.tp_end_offsets = {}

        self.stream_inbound_time = {}
        Service.__init__(self, **kwargs)

    def secs_since(self, start_time: float) -> float:
        """Given timestamp start, return number of seconds since that time."""
        return self.time() - start_time

    def ms_since(self, start_time: float) -> float:
        """Given timestamp start, return number of ms since that time."""
        return self.secs_to_ms(self.secs_since(start_time))

    def secs_to_ms(self, timestamp: float) -> float:
        """Convert seconds to milliseconds."""
        return timestamp * 1000.

    @Service.task
    async def _sampler(self) -> None:
        prev_message_total = self.messages_received_total
        prev_event_total = self.events_total

        async for sleep_time in self.itertimer(1.0, name='Monitor.sampler'):
            prev_event_total, prev_message_total = self._sample(
                prev_event_total, prev_message_total)

    def _sample(self,
                prev_event_total: int,
                prev_message_total: int) -> Tuple[int, int]:
        # Update average event runtime.
        if self.events_runtime:
            self.events_runtime_avg = median(self.events_runtime)

        # Update events/s
        self.events_s, prev_event_total = (
            self.events_total - prev_event_total,
            self.events_total,
        )

        # Update messages/s
        self.messages_s, prev_message_total = (
            self.messages_received_total - prev_message_total,
            self.messages_received_total)

        if self.rebalance_return_latency:
            self.rebalance_return_avg = median(self.rebalance_return_latency)

        if self.rebalance_end_latency:
            self.rebalance_end_avg = median(self.rebalance_end_latency)

        if self.http_response_latency:
            self.http_response_latency_avg = median(
                self.http_response_latency)

        return prev_event_total, prev_message_total

    def asdict(self) -> Mapping:
        """Return monitor state as dictionary."""
        return {
            'messages_active': self.messages_active,
            'messages_received_total': self.messages_received_total,
            'messages_sent': self.messages_sent,
            'messages_sent_by_topic': self.messages_sent_by_topic,
            'messages_s': self.messages_s,
            'messages_received_by_topic': self.messages_received_by_topic,
            'events_active': self.events_active,
            'events_total': self.events_total,
            'events_s': self.events_s,
            'events_runtime_avg': self.events_runtime_avg,
            'events_by_task': self._events_by_task_dict(),
            'events_by_stream': self._events_by_stream_dict(),
            'commit_latency': self.commit_latency,
            'send_latency': self.send_latency,
            'send_errors': self.send_errors,
            'assignment_latency': self.assignment_latency,
            'assignments_completed': self.assignments_completed,
            'assignments_failed': self.assignments_failed,
            'topic_buffer_full': self._topic_buffer_full_dict(),
            'tables': {
                name: table.asdict() for name, table in self.tables.items()
            },
            'metric_counts': self._metric_counts_dict(),
            'topic_committed_offsets': self._tp_committed_offsets_dict(),
            'topic_read_offsets': self._tp_read_offsets_dict(),
            'topic_end_offsets': self._tp_end_offsets_dict(),
            'rebalances': self.rebalances,
            'rebalance_return_latency': self.rebalance_return_latency,
            'rebalance_end_latency': self.rebalance_end_latency,
            'rebalance_return_avg': self.rebalance_return_avg,
            'rebalance_end_avg': self.rebalance_end_avg,
            'http_response_codes': self._http_response_codes_dict(),
            'http_response_latency': self.http_response_latency,
            'http_response_latency_avg': self.http_response_latency_avg,
        }

    def _events_by_stream_dict(self) -> MutableMapping[str, int]:
        return {label(stream): count
                for stream, count in self.events_by_stream.items()}

    def _events_by_task_dict(self) -> MutableMapping[str, int]:
        return {label(task): count
                for task, count in self.events_by_task.items()}

    def _topic_buffer_full_dict(self) -> MutableMapping[str, int]:
        return {label(topic): count
                for topic, count in self.topic_buffer_full.items()}

    def _metric_counts_dict(self) -> MutableMapping[str, int]:
        return {key: count for key, count in self.metric_counts.items()}

    def _http_response_codes_dict(self) -> MutableMapping[int, int]:
        return {int(code): count
                for code, count in self.http_response_codes.items()}

    def _tp_committed_offsets_dict(self) -> TPOffsetDict:
        return self._tp_offsets_as_dict(self.tp_committed_offsets)

    def _tp_read_offsets_dict(self) -> TPOffsetDict:
        return self._tp_offsets_as_dict(self.tp_read_offsets)

    def _tp_end_offsets_dict(self) -> TPOffsetDict:
        return self._tp_offsets_as_dict(self.tp_end_offsets)

    @classmethod
    def _tp_offsets_as_dict(cls, tp_offsets: TPOffsetMapping) -> TPOffsetDict:
        topic_partition_offsets: TPOffsetDict = {}
        for tp, offset in tp_offsets.items():
            partition_offsets = topic_partition_offsets.get(tp.topic) or {}
            partition_offsets[tp.partition] = offset
            topic_partition_offsets[tp.topic] = partition_offsets
        return topic_partition_offsets

    def on_message_in(self, tp: TP, offset: int, message: Message) -> None:
        """Call before message is delegated to streams."""
        # WARNING: Sensors must never keep a reference to the Message,
        #          as this means the message won't go out of scope!
        self.messages_received_total += 1
        self.messages_active += 1
        self.messages_received_by_topic[tp.topic] += 1
        self.tp_read_offsets[tp] = offset
        message.time_in = self.time()

    def on_stream_event_in(self, tp: TP, offset: int, stream: StreamT,
                           event: EventT) -> Optional[Dict]:
        """Call when stream starts processing an event."""
        self.events_total += 1
        self.events_by_stream[str(stream)] += 1
        self.events_by_task[str(stream.task_owner)] += 1
        self.events_active += 1
        return {
            'time_in': self.time(),
            'time_out': None,
            'time_total': None,
        }

    def on_stream_event_out(self, tp: TP, offset: int, stream: StreamT,
                            event: EventT, state: Dict = None) -> None:
        """Call when stream is done processing an event."""
        if state is not None:
            time_out = self.time()
            time_in = state['time_in']
            time_total = time_out - time_in
            self.events_active -= 1
            state.update(
                time_out=time_out,
                time_total=time_total,
            )
            deque_pushpopmax(
                self.events_runtime, time_total, self.max_avg_history)

    def on_topic_buffer_full(self, topic: TopicT) -> None:
        """Call when conductor topic buffer is full and has to wait."""
        self.topic_buffer_full[topic] += 1

    def on_message_out(self,
                       tp: TP,
                       offset: int,
                       message: Message) -> None:
        """Call when message is fully acknowledged and can be committed."""
        self.messages_active -= 1
        time_out = message.time_out = self.time()
        time_in = message.time_in
        if time_in is not None:
            message.time_total = time_out - time_in

    def on_table_get(self, table: CollectionT, key: Any) -> None:
        """Call when value in table is retrieved."""
        self._table_or_create(table).keys_retrieved += 1

    def on_table_set(self, table: CollectionT, key: Any, value: Any) -> None:
        """Call when new value for key in table is set."""
        self._table_or_create(table).keys_updated += 1

    def on_table_del(self, table: CollectionT, key: Any) -> None:
        """Call when key in a table is deleted."""
        self._table_or_create(table).keys_deleted += 1

    def _table_or_create(self, table: CollectionT) -> TableState:
        try:
            return self.tables[table.name]
        except KeyError:
            state = self.tables[table.name] = TableState(table)
            return state

    def on_commit_initiated(self, consumer: ConsumerT) -> Any:
        """Consumer is about to commit topic offset."""
        return self.time()

    def on_commit_completed(self, consumer: ConsumerT, state: Any) -> None:
        """Call when consumer commit offset operation completed."""
        latency = self.time() - cast(float, state)
        deque_pushpopmax(
            self.commit_latency,
            latency,
            self.max_commit_latency_history,
        )

    def on_send_initiated(self, producer: ProducerT, topic: str,
                          message: PendingMessage,
                          keysize: int, valsize: int) -> Any:
        """Call when message added to producer buffer."""
        self.messages_sent += 1
        self.messages_sent_by_topic[topic] += 1
        return self.time()

    def on_send_completed(self,
                          producer: ProducerT,
                          state: Any,
                          metadata: RecordMetadata) -> None:
        """Call when producer finished sending message."""
        latency = self.time() - cast(float, state)
        deque_pushpopmax(
            self.send_latency, latency, self.max_send_latency_history)

    def on_send_error(self,
                      producer: ProducerT,
                      exc: BaseException,
                      state: Any) -> None:
        """Call when producer was unable to publish message."""
        self.send_errors += 1

    def count(self, metric_name: str, count: int = 1) -> None:
        """Count metric by name."""
        self.metric_counts[metric_name] += count

    def on_tp_commit(self, tp_offsets: TPOffsetMapping) -> None:
        """Call when offset in topic partition is committed."""
        self.tp_committed_offsets.update(tp_offsets)

    def track_tp_end_offset(self, tp: TP, offset: int) -> None:
        """Track new topic partition end offset for monitoring lags."""
        self.tp_end_offsets[tp] = offset

    def on_assignment_start(self,
                            assignor: PartitionAssignorT) -> Dict:
        """Partition assignor is starting to assign partitions."""
        return {'time_start': self.time()}

    def on_assignment_error(self,
                            assignor: PartitionAssignorT,
                            state: Dict,
                            exc: BaseException) -> None:
        """Partition assignor did not complete assignor due to error."""
        time_total = self.time() - state['time_start']
        deque_pushpopmax(
            self.assignment_latency, time_total,
            self.max_assignment_latency_history)
        self.assignments_failed += 1

    def on_assignment_completed(self,
                                assignor: PartitionAssignorT,
                                state: Dict) -> None:
        """Partition assignor completed assignment."""
        time_total = self.time() - state['time_start']
        deque_pushpopmax(
            self.assignment_latency, time_total,
            self.max_assignment_latency_history)
        self.assignments_completed += 1

    def on_rebalance_start(self, app: AppT) -> Dict:
        """Cluster rebalance in progress."""
        self.rebalances = app.rebalancing_count
        return {'time_start': self.time()}

    def on_rebalance_return(self, app: AppT, state: Dict) -> None:
        """Consumer replied assignment is done to broker."""
        time_start = state['time_start']
        time_return = self.time()
        latency_return = time_return - time_start
        state.update(
            time_return=time_return,
            latency_return=latency_return,
        )
        deque_pushpopmax(
            self.rebalance_return_latency,
            latency_return,
            self.max_avg_history)

    def on_rebalance_end(self, app: AppT, state: Dict) -> None:
        """Cluster rebalance fully completed (including recovery)."""
        time_start = state['time_start']
        time_end = self.time()
        latency_end = time_end - time_start
        state.update(
            time_end=time_end,
            latency_end=latency_end,
        )
        deque_pushpopmax(
            self.rebalance_end_latency, latency_end, self.max_avg_history)

    def on_web_request_start(self, app: AppT, request: web.Request, *,
                             view: web.View = None) -> Dict:
        """Web server started working on request."""
        return {'time_start': self.time()}

    def on_web_request_end(self,
                           app: AppT,
                           request: web.Request,
                           response: Optional[web.Response],
                           state: Dict,
                           *,
                           view: web.View = None) -> None:
        """Web server finished working on request."""
        status_code = HTTPStatus(response.status if response else 500)
        time_start = state['time_start']
        time_end = self.time()
        latency_end = time_end - time_start
        state.update(
            time_end=time_end,
            latency_end=latency_end,
            status_code=status_code,
        )
        deque_pushpopmax(
            self.http_response_latency, latency_end, self.max_avg_history)
        self.http_response_codes[status_code] += 1
